"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import pytest

from salt.cloud.clouds import linode


@pytest.fixture
def configure_loader_modules():
    return {linode: {}}


# _validate_name tests


def test_validate_name_first_character_invalid():
    """
    Tests when name starts with an invalid character.
    """
    # Test when name begins with a hyphen
    assert linode._validate_name("-foo") is False

    # Test when name begins with an underscore
    assert linode._validate_name("_foo") is False


def test_validate_name_last_character_invalid():
    """
    Tests when name ends with an invalid character.
    """
    # Test when name ends with a hyphen
    assert linode._validate_name("foo-") is False

    # Test when name ends with an underscore
    assert linode._validate_name("foo_") is False


def test_validate_name_too_short():
    """
    Tests when name has less than three letters.
    """
    # Test when name is an empty string
    assert linode._validate_name("") is False

    # Test when name is two letters long
    assert linode._validate_name("ab") is False

    # Test when name is three letters long (valid)
    assert linode._validate_name("abc") is True


def test_validate_name_too_long():
    """
    Tests when name has more than 48 letters.
    """
    long_name = "1111-2222-3333-4444-5555-6666-7777-8888-9999-111"
    # Test when name is 48 letters long (valid)
    assert len(long_name) == 48
    assert linode._validate_name(long_name) is True

    # Test when name is more than 48 letters long
    long_name += "1"
    assert len(long_name) == 49
    assert linode._validate_name(long_name) is False


def test_validate_name_invalid_characters():
    """
    Tests when name contains invalid characters.
    """
    # Test when name contains an invalid character
    assert linode._validate_name("foo;bar") is False

    # Test when name contains non-ascii letters
    assert linode._validate_name("fooàààààbar") is False

    # Test when name contains spaces
    assert linode._validate_name("foo bar") is False


def test_validate_name_valid_characters():
    """
    Tests when name contains valid characters.
    """
    # Test when name contains letters and numbers
    assert linode._validate_name("foo123bar") is True

    # Test when name contains hyphens
    assert linode._validate_name("foo-bar") is True

    # Test when name contains underscores
    assert linode._validate_name("foo_bar") is True

    # Test when name start and end with numbers
    assert linode._validate_name("1foo") is True
    assert linode._validate_name("foo0") is True



"""
Test States
===========

Provide test case states that enable easy testing of things to do with state
calls, e.g. running, calling, logging, output filtering etc.

.. code-block:: yaml

    always-passes-with-any-kwarg:
      test.nop:
        - name: foo
        - something: else
        - foo: bar

    always-passes:
      test.succeed_without_changes:
        - name: foo

    always-fails:
      test.fail_without_changes:
        - name: foo

    always-changes-and-succeeds:
      test.succeed_with_changes:
        - name: foo

    always-changes-and-fails:
      test.fail_with_changes:
        - name: foo

    my-custom-combo:
      test.configurable_test_state:
        - name: foo
        - changes: True
        - result: False
        - comment: bar.baz
        - warnings: A warning

    is-pillar-foo-present-and-bar-is-int:
      test.check_pillar:
        - present:
            - foo
        - integer:
            - bar

You may also use these states for controlled failure in state definitions, for example if certain conditions in
pillar or grains do not apply. The following state definition will fail with a message "OS not supported!" when
`grains['os']` is neither Ubuntu nor CentOS:

.. code-block:: jinja

    {% if grains['os'] in ['Ubuntu', 'CentOS'] %}

    # Your state definitions go here

    {% else %}
    failure:
      test.fail_without_changes:
        - name: "OS not supported!"
        - failhard: True
    {% endif %}

"""

import random

import salt.utils.data
from salt.exceptions import SaltInvocationError
from salt.state import _gen_tag


def nop(name, **kwargs):
    """
    .. versionadded:: 2015.8.1

    A no-op state that does nothing. Useful in conjunction with the ``use``
    requisite, or in templates which could otherwise be empty due to jinja
    rendering.

    name
        A unique string to serve as the state's ID
    """
    return succeed_without_changes(name)


def succeed_without_changes(name, **kwargs):  # pylint: disable=unused-argument
    """
    .. versionadded:: 2014.7.0

    Returns successful

    name
        A unique string to serve as the state's ID
    """
    comment = kwargs.get("comment", "Success!")

    ret = {"name": name, "changes": {}, "result": True, "comment": comment}
    return ret


def fail_without_changes(name, **kwargs):  # pylint: disable=unused-argument
    """
    .. versionadded:: 2014.7.0

    Returns failure

    name
        A unique string to serve as the state's ID
    """
    comment = kwargs.get("comment", "Failure!")

    ret = {"name": name, "changes": {}, "result": False, "comment": comment}

    if __opts__["test"]:
        ret["result"] = False
        ret["comment"] = "If we weren't testing, this would be a failure!"

    return ret


def succeed_with_changes(name, **kwargs):  # pylint: disable=unused-argument
    """
    .. versionadded:: 2014.7.0

    Returns ``True`` with an non-empty ``changes`` dictionary. Useful for
    testing requisites.

    name
        A unique string to serve as the state's ID
    """
    comment = kwargs.get("comment", "Success!")

    ret = {"name": name, "changes": {}, "result": True, "comment": comment}

    # Following the docs as written here
    # https://docs.saltproject.io/ref/states/writing.html#return-data
    ret["changes"] = {
        "testing": {"old": "Unchanged", "new": "Something pretended to change"}
    }

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "If we weren't testing, this would be successful with changes"

    return ret


def fail_with_changes(name, **kwargs):  # pylint: disable=unused-argument
    """
    .. versionadded:: 2014.7.0

    Returns ``False`` with an non-empty ``changes`` dictionary. Useful for
    testing requisites.

    name
        A unique string to serve as the state's ID
    """
    comment = kwargs.get("comment", "Failure!")

    ret = {"name": name, "changes": {}, "result": False, "comment": comment}

    # Following the docs as written here
    # https://docs.saltproject.io/ref/states/writing.html#return-data
    ret["changes"] = {
        "testing": {"old": "Unchanged", "new": "Something pretended to change"}
    }

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "If we weren't testing, this would be failed with changes"

    return ret


def configurable_test_state(name, changes=True, result=True, comment="", warnings=None):
    """
    .. versionadded:: 2014.7.0

    A configurable test state which allows for more control over the return
    data

    name
        A unique string to serve as the state's ID

    changes : True
        Controls whether or not the state reports that there were changes.
        There are three supported values for this argument:

        - If ``True``, the state will report changes
        - If ``False``, the state will report no changes
        - If ``"Random"``, the state will randomly report either changes or no
          changes.

    result : True
        Controls the result for for the state. Like ``changes``, there are
        three supported values for this argument:

        - If ``True``, the state will report a ``True`` result
        - If ``False``, the state will report a ``False`` result
        - If ``"Random"``, the state will randomly report either ``True``

        .. note::
            The result will be reported as ``None`` if *all* of the following
            are true:

            1. The state is being run in test mode (i.e. ``test=True`` on the
            CLI)

            2. ``result`` is ``True`` (either explicitly, or via being set to
               ``"Random"``)

            3. ``changes`` is ``True`` (either explicitly, or via being set to
               ``"Random"``)

    comment : ""
        Comment field field for the state. By default, this is an empty string.

    warnings
        A string (or a list of strings) to fill the warnings field with.
        Default is None

        .. versionadded:: 3000
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": comment}
    change_data = {
        "testing": {"old": "Unchanged", "new": "Something pretended to change"}
    }

    if changes is True:
        # If changes is True we place our dummy change dictionary into it.
        # Following the docs as written here
        # https://docs.saltproject.io/ref/states/writing.html#return-data
        ret["changes"] = change_data
    elif changes is False:
        # Don't modify changes from the "ret" dict set above
        pass
    else:
        if str(changes).lower() == "random":
            if random.choice((True, False)):
                # Following the docs as written here
                # https://docs.saltproject.io/ref/states/writing.html#return-data
                ret["changes"] = change_data
        else:
            err = (
                "You have specified the state option 'changes' with "
                "invalid arguments. It must be either "
                "True, False, or 'Random'"
            )
            raise SaltInvocationError(err)

    if isinstance(result, bool):
        ret["result"] = result
    else:
        if str(result).lower() == "random":
            ret["result"] = random.choice((True, False))
        else:
            raise SaltInvocationError(
                "You have specified the state option "
                "'result' with invalid arguments. It must "
                "be either True, False, or "
                "'Random'"
            )

    if warnings is None:
        pass
    elif isinstance(warnings, str):
        ret["warnings"] = [warnings]
    elif isinstance(warnings, list):
        ret["warnings"] = warnings
    else:
        raise SaltInvocationError(
            "You have specified the state option "
            "'warnings' with invalid arguments. It must "
            "be a string or a list of strings"
        )

    if __opts__["test"]:
        ret["result"] = True if changes is False else None
        ret["comment"] = "This is a test" if not comment else comment

    return ret


def show_notification(name, text=None, **kwargs):
    """
    .. versionadded:: 2015.8.0

    Simple notification using text argument.

    name
        A unique string to serve as the state's ID

    text
        Text to return in the comment field
    """

    if not text:
        raise SaltInvocationError("Missing required argument text.")

    ret = {"name": name, "changes": {}, "result": True, "comment": text}

    return ret


def mod_watch(name, sfun=None, **kwargs):
    """
    Call this function via a watch statement

    .. versionadded:: 2014.7.0

    Any parameters in the state return dictionary can be customized by adding
    the keywords ``result``, ``comment``, and ``changes``.

    .. code-block:: yaml

        this_state_will_return_changes:
          test.succeed_with_changes

        this_state_will_NOT_return_changes:
          test.succeed_without_changes

        this_state_is_watching_another_state:
          test.succeed_without_changes:
            - comment: 'This is a custom comment'
            - watch:
              - test: this_state_will_return_changes
              - test: this_state_will_NOT_return_changes

        this_state_is_also_watching_another_state:
          test.succeed_without_changes:
            - watch:
              - test: this_state_will_NOT_return_changes
    """
    has_changes = []
    if "__reqs__" in __low__:
        for req in __low__["__reqs__"]["watch"]:
            tag = _gen_tag(req)
            if __running__[tag]["changes"]:
                has_changes.append("{state}: {__id__}".format(**req))

    ret = {
        "name": name,
        "result": kwargs.get("result", True),
        "comment": kwargs.get("comment", "Watch statement fired."),
        "changes": kwargs.get("changes", {"Requisites with changes": has_changes}),
    }
    return ret


def _check_key_type(key_str, key_type=None):
    """
    Helper function to get pillar[key_str] and
    check if its type is key_type

    Returns None if the pillar key is missing.
    If present True or False depending on match
    of the values type.

    Can't check for None.
    """
    value = __salt__["pillar.get"](key_str, None)
    if value is None:
        return None
    elif key_type is not None and not isinstance(value, key_type):
        return False
    else:
        return True


def _if_str_then_list(listing):
    """
    Checks if its argument is a list or a str.
    A str will be turned into a list with the
    str as its only element.
    """
    if isinstance(listing, str):
        return [salt.utils.stringutils.to_unicode(listing)]
    elif not isinstance(listing, list):
        raise TypeError
    return salt.utils.data.decode_list(listing)


def check_pillar(
    name,
    present=None,
    boolean=None,
    integer=None,
    string=None,
    listing=None,
    dictionary=None,
    verbose=False,
):
    """
    Checks the presence and, optionally, the type of given keys in Pillar

    Supported kwargs for types are:
    - boolean (bool)
    - integer (int)
    - string (str)
    - listing (list)
    - dictionary (dict)

    Checking for None type pillars is not implemented yet.

    .. code-block:: yaml

        is-pillar-foo-present-and-bar-is-int:
          test.check_pillar:
            - present:
                - foo
            - integer:
                - bar
    """
    if not (present or boolean or integer or string or listing or dictionary):
        raise SaltInvocationError("Missing required argument text.")

    present = present or []
    boolean = boolean or []
    integer = integer or []
    string = string or []
    listing = listing or []
    dictionary = dictionary or []

    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    checks = {}
    fine = {}
    failed = {}
    # for those we don't check the type:
    present = _if_str_then_list(present)
    checks[None] = present
    # those should be bool:
    boolean = _if_str_then_list(boolean)
    checks[bool] = boolean
    # those should be int:

    # those should be integer:
    integer = _if_str_then_list(integer)
    checks[int] = integer
    # those should be str:
    string = _if_str_then_list(string)
    checks[(str,)] = string
    # those should be list:
    listing = _if_str_then_list(listing)
    checks[list] = listing
    # those should be dict:
    dictionary = _if_str_then_list(dictionary)
    checks[dict] = dictionary

    for key_type, keys in checks.items():
        for key in keys:
            result = _check_key_type(key, key_type)
            if result is None:
                failed[key] = None
                ret["result"] = False
            elif not result:
                failed[key] = key_type
                ret["result"] = False
            elif verbose:
                fine[key] = key_type

    for key, key_type in failed.items():
        comment = 'Pillar key "{}" '.format(key)
        if key_type is None:
            comment += "is missing.\n"
        else:
            comment += "is not {}.\n".format(key_type)
        ret["comment"] += comment

    if verbose and fine:
        comment = "Those keys passed the check:\n"
        for key, key_type in fine.items():
            comment += "- {} ({})\n".format(key, key_type)
        ret["comment"] += comment

    return ret



"""
    Tests for salt.modules.boto3_elasticsearch
"""
import datetime
import random
import string
import textwrap

import pytest

import salt.loader
import salt.modules.boto3_elasticsearch as boto3_elasticsearch
from salt.utils.versions import Version
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

try:
    import boto3
    from botocore.exceptions import ClientError

    HAS_BOTO3 = True
except ImportError:
    HAS_BOTO3 = False

# the boto3_elasticsearch module relies on the connect_to_region() method
# which was added in boto 2.8.0
# https://github.com/boto/boto/commit/33ac26b416fbb48a60602542b4ce15dcc7029f12
REQUIRED_BOTO3_VERSION = "1.2.1"


def __virtual__():
    """
    Returns True/False boolean depending on if Boto3 is installed and correct
    version.
    """
    if not HAS_BOTO3:
        return False
    if Version(boto3.__version__) < Version(REQUIRED_BOTO3_VERSION):
        return (
            False,
            "The boto3 module must be greater or equal to version {}".format(
                REQUIRED_BOTO3_VERSION
            ),
        )
    return True


REGION = "us-east-1"
ACCESS_KEY = "GKTADJGHEIQSXMKKRBJ08H"
SECRET_KEY = "askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs"
CONN_PARAMETERS = {
    "region": REGION,
    "key": ACCESS_KEY,
    "keyid": SECRET_KEY,
    "profile": {},
}
ERROR_MESSAGE = (
    "An error occurred ({}) when calling the {} operation: Test-defined error"
)
ERROR_CONTENT = {"Error": {"Code": 101, "Message": "Test-defined error"}}
NOT_FOUND_ERROR = ClientError(
    {"Error": {"Code": "ResourceNotFoundException", "Message": "Test-defined error"}},
    "msg",
)
DOMAIN_RET = {
    "DomainId": "accountno/testdomain",
    "DomainName": "testdomain",
    "ARN": "arn:aws:es:region:accountno:domain/testdomain",
    "Created": True,
    "Deleted": False,
    "Endpoints": {"vpc": "vpc-testdomain-1234567890.region.es.amazonaws.com"},
    "Processing": False,
    "UpgradeProcessing": False,
    "ElasticsearchVersion": "6.3",
    "ElasticsearchClusterConfig": {
        "InstanceType": "t2.medium.elasticsearch",
        "InstanceCount": 1,
        "DedicatedMasterEnabled": False,
        "ZoneAwarenessEnabled": False,
    },
    "EBSOptions": {
        "EBSEnabled": True,
        "VolumeType": "gp2",
        "VolumeSize": 123,
        "Iops": 12,
    },
    "AccessPolicies": textwrap.dedent(
        """
        {"Version":"2012-10-17","Statement":[{"Effect":"Allow",
        "Principal":{"AWS":"*"},"Action":"es:*",
        "Resource":"arn:aws:es:region:accountno:domain/testdomain/*"}]}"""
    ),
    "SnapshotOptions": {"AutomatedSnapshotStartHour": 1},
    "VPCOptions": {
        "VPCId": "vpc-12345678",
        "SubnetIds": ["subnet-deadbeef"],
        "AvailabilityZones": ["regiona"],
        "SecurityGroupIds": ["sg-87654321"],
    },
    "CognitoOptions": {"Enabled": False},
    "EncryptionAtRestOptions": {"Enabled": False},
    "NodeToNodeEncryptionOptions": {"Enabled": False},
    "AdvancedOptions": {"rest.action.multi.allow_explicit_index": "true"},
    "ServiceSoftwareOptions": {
        "CurrentVersion": "R20190221-P1",
        "NewVersion": "R20190418",
        "UpdateAvailable": True,
        "Cancellable": False,
        "UpdateStatus": "ELIGIBLE",
        "Description": (
            "A newer release R20190418 is available. This release "
            "will be automatically deployed after somedate"
        ),
        "AutomatedUpdateDate": None,
    },
}


@pytest.mark.skipif(HAS_BOTO3 is False, reason="The boto module must be installed.")
@pytest.mark.skipif(
    Version(boto3.__version__) < Version(REQUIRED_BOTO3_VERSION),
    reason="The boto3 module must be greater or equal to version {}".format(
        REQUIRED_BOTO3_VERSION
    ),
)
class Boto3ElasticsearchTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.modules.boto3_elasticsearch module
    """

    conn = None

    def setup_loader_modules(self):
        self.opts = salt.config.DEFAULT_MINION_OPTS.copy()
        utils = salt.loader.utils(
            self.opts,
            whitelist=["boto3", "args", "systemd", "path", "platform"],
            context={},
        )
        return {boto3_elasticsearch: {"__utils__": utils}}

    def setUp(self):
        super().setUp()
        boto3_elasticsearch.__init__(self.opts)
        del self.opts

        # Set up MagicMock to replace the boto3 session
        # connections keep getting cached from prior tests, can't find the
        # correct context object to clear it. So randomize the cache key, to prevent any
        # cache hits
        CONN_PARAMETERS["key"] = "".join(
            random.choice(string.ascii_lowercase + string.digits) for _ in range(50)
        )

        self.conn = MagicMock()
        self.addCleanup(delattr, self, "conn")
        self.patcher = patch("boto3.session.Session")
        self.addCleanup(self.patcher.stop)
        self.addCleanup(delattr, self, "patcher")
        mock_session = self.patcher.start()
        session_instance = mock_session.return_value
        session_instance.configure_mock(client=MagicMock(return_value=self.conn))
        self.paginator = MagicMock()
        self.addCleanup(delattr, self, "paginator")
        self.conn.configure_mock(get_paginator=MagicMock(return_value=self.paginator))

    def test_describe_elasticsearch_domain_positive(self):
        """
        Test that when describing a domain when the domain actually exists,
        the .exists method returns a dict with 'result': True
        and 'response' with the domain status information.
        """
        # The patch below is not neccesary per se,
        # as .exists returns positive as long as no exception is raised.
        with patch.object(
            self.conn,
            "describe_elasticsearch_domain",
            return_value={"DomainStatus": DOMAIN_RET},
        ):
            self.assertEqual(
                boto3_elasticsearch.describe_elasticsearch_domain(
                    domain_name="testdomain", **CONN_PARAMETERS
                ),
                {"result": True, "response": DOMAIN_RET},
            )

    def test_describe_elasticsearch_domain_error(self):
        """
        Test that when describing a domain when the domain does not exist,
        the .exists method returns a dict with 'result': False
        and 'error' with boto's ResourceNotFoundException.
        """
        with patch.object(
            self.conn, "describe_elasticsearch_domain", side_effect=NOT_FOUND_ERROR
        ):
            result = boto3_elasticsearch.describe_elasticsearch_domain(
                domain_name="testdomain", **CONN_PARAMETERS
            )
            self.assertEqual(
                result.get("error", ""),
                ERROR_MESSAGE.format("ResourceNotFoundException", "msg"),
            )
            self.assertFalse(result["result"])

    def test_create_elasticsearch_domain_positive(self):
        """
        Test that when creating a domain, and it succeeds,
        the .create method returns a dict with 'result': True
        and 'response' with the newly created domain's status information.
        """
        with patch.object(
            self.conn,
            "create_elasticsearch_domain",
            return_value={"DomainStatus": DOMAIN_RET},
        ):
            kwargs = {
                "elasticsearch_version": DOMAIN_RET["ElasticsearchVersion"],
                "elasticsearch_cluster_config": DOMAIN_RET[
                    "ElasticsearchClusterConfig"
                ],
                "ebs_options": DOMAIN_RET["EBSOptions"],
                "access_policies": DOMAIN_RET["AccessPolicies"],
                "snapshot_options": DOMAIN_RET["SnapshotOptions"],
                "vpc_options": DOMAIN_RET["VPCOptions"],
                "cognito_options": DOMAIN_RET["CognitoOptions"],
                "encryption_at_rest_options": DOMAIN_RET["EncryptionAtRestOptions"],
                "advanced_options": DOMAIN_RET["AdvancedOptions"],
            }
            kwargs.update(CONN_PARAMETERS)
            self.assertEqual(
                boto3_elasticsearch.create_elasticsearch_domain(
                    domain_name="testdomain", **kwargs
                ),
                {"result": True, "response": DOMAIN_RET},
            )

    def test_create_elasticsearch_domain_error(self):
        """
        Test that when creating a domain, and boto3 returns an error,
        the .create method returns a dict with 'result': False
        and 'error' with the error reported by boto3.
        """
        with patch.object(
            self.conn,
            "create_elasticsearch_domain",
            side_effect=ClientError(ERROR_CONTENT, "create_domain"),
        ):
            kwargs = {
                "elasticsearch_version": DOMAIN_RET["ElasticsearchVersion"],
                "elasticsearch_cluster_config": DOMAIN_RET[
                    "ElasticsearchClusterConfig"
                ],
                "ebs_options": DOMAIN_RET["EBSOptions"],
                "access_policies": DOMAIN_RET["AccessPolicies"],
                "snapshot_options": DOMAIN_RET["SnapshotOptions"],
                "vpc_options": DOMAIN_RET["VPCOptions"],
                "cognito_options": DOMAIN_RET["CognitoOptions"],
                "encryption_at_rest_options": DOMAIN_RET["EncryptionAtRestOptions"],
                "advanced_options": DOMAIN_RET["AdvancedOptions"],
            }
            kwargs.update(CONN_PARAMETERS)
            result = boto3_elasticsearch.create_elasticsearch_domain(
                "testdomain", **kwargs
            )
            self.assertEqual(
                result.get("error", ""), ERROR_MESSAGE.format(101, "create_domain")
            )

    def test_delete_domain_positive(self):
        """
        Test that when deleting a domain, and it succeeds,
        the .delete method returns {'result': True}.
        """
        with patch.object(self.conn, "delete_elasticsearch_domain"):
            self.assertEqual(
                boto3_elasticsearch.delete_elasticsearch_domain(
                    "testdomain", **CONN_PARAMETERS
                ),
                {"result": True},
            )

    def test_delete_domain_error(self):
        """
        Test that when deleting a domain, and boto3 returns an error,
        the .delete method returns {'result': False, 'error' :'the error'}.
        """
        with patch.object(
            self.conn,
            "delete_elasticsearch_domain",
            side_effect=ClientError(ERROR_CONTENT, "delete_domain"),
        ):
            result = boto3_elasticsearch.delete_elasticsearch_domain(
                "testdomain", **CONN_PARAMETERS
            )
            self.assertFalse(result["result"])
            self.assertEqual(
                result.get("error", ""), ERROR_MESSAGE.format(101, "delete_domain")
            )

    def test_update_domain_positive(self):
        """
        Test that when updating a domain succeeds, the .update method returns {'result': True}.
        """
        with patch.object(
            self.conn,
            "update_elasticsearch_domain_config",
            return_value={"DomainConfig": DOMAIN_RET},
        ):
            kwargs = {
                "elasticsearch_cluster_config": DOMAIN_RET[
                    "ElasticsearchClusterConfig"
                ],
                "ebs_options": DOMAIN_RET["EBSOptions"],
                "snapshot_options": DOMAIN_RET["SnapshotOptions"],
                "vpc_options": DOMAIN_RET["VPCOptions"],
                "cognito_options": DOMAIN_RET["CognitoOptions"],
                "advanced_options": DOMAIN_RET["AdvancedOptions"],
                "access_policies": DOMAIN_RET["AccessPolicies"],
                "log_publishing_options": {},
            }

            kwargs.update(CONN_PARAMETERS)
            self.assertEqual(
                boto3_elasticsearch.update_elasticsearch_domain_config(
                    "testdomain", **kwargs
                ),
                {"result": True, "response": DOMAIN_RET},
            )

    def test_update_domain_error(self):
        """
        Test that when updating a domain fails, and boto3 returns an error,
        the .update method returns the error.
        """
        with patch.object(
            self.conn,
            "update_elasticsearch_domain_config",
            side_effect=ClientError(ERROR_CONTENT, "update_domain"),
        ):
            kwargs = {
                "elasticsearch_cluster_config": DOMAIN_RET[
                    "ElasticsearchClusterConfig"
                ],
                "ebs_options": DOMAIN_RET["EBSOptions"],
                "snapshot_options": DOMAIN_RET["SnapshotOptions"],
                "vpc_options": DOMAIN_RET["VPCOptions"],
                "cognito_options": DOMAIN_RET["CognitoOptions"],
                "advanced_options": DOMAIN_RET["AdvancedOptions"],
                "access_policies": DOMAIN_RET["AccessPolicies"],
                "log_publishing_options": {},
            }
            kwargs.update(CONN_PARAMETERS)
            result = boto3_elasticsearch.update_elasticsearch_domain_config(
                "testdomain", **kwargs
            )
            self.assertEqual(
                result.get("error", ""), ERROR_MESSAGE.format(101, "update_domain")
            )

    def test_add_tags_positive(self):
        """
        Test that when adding tags is successful, the .add_tags method returns {'result': True}.
        """
        with patch.object(
            self.conn,
            "describe_elasticsearch_domain",
            return_value={"DomainStatus": DOMAIN_RET},
        ):
            self.assertEqual(
                boto3_elasticsearch.add_tags(
                    "testdomain", tags={"foo": "bar", "baz": "qux"}, **CONN_PARAMETERS
                ),
                {"result": True},
            )

    def test_add_tags_default(self):
        """
        Test that when tags are not provided, no error is raised.
        """
        with patch.object(
            self.conn,
            "describe_elasticsearch_domain",
            return_value={"DomainStatus": DOMAIN_RET},
        ):
            self.assertEqual(
                boto3_elasticsearch.add_tags("testdomain", **CONN_PARAMETERS),
                {"result": True},
            )

    def test_add_tags_error(self):
        """
        Test that when adding tags fails, and boto3 returns an error,
        the .add_tags function returns {'tagged': False, 'error': 'the error'}.
        """
        with patch.object(
            self.conn, "add_tags", side_effect=ClientError(ERROR_CONTENT, "add_tags")
        ), patch.object(
            self.conn,
            "describe_elasticsearch_domain",
            return_value={"DomainStatus": DOMAIN_RET},
        ):
            result = boto3_elasticsearch.add_tags(
                "testdomain", tags={"foo": "bar", "baz": "qux"}, **CONN_PARAMETERS
            )
            self.assertFalse(result["result"])
            self.assertEqual(
                result.get("error", ""), ERROR_MESSAGE.format(101, "add_tags")
            )

    def test_remove_tags_positive(self):
        """
        Test that when removing tags is successful, the .remove_tags method returns {'tagged': True}.
        """
        with patch.object(
            self.conn,
            "describe_elasticsearch_domain",
            return_value={"DomainStatus": DOMAIN_RET},
        ):
            self.assertEqual(
                boto3_elasticsearch.remove_tags(
                    tag_keys=["foo", "bar"], domain_name="testdomain", **CONN_PARAMETERS
                ),
                {"result": True},
            )

    def test_remove_tag_error(self):
        """
        Test that when removing tags fails, and boto3 returns an error,
        the .remove_tags method returns {'tagged': False, 'error': 'the error'}.
        """
        with patch.object(
            self.conn,
            "remove_tags",
            side_effect=ClientError(ERROR_CONTENT, "remove_tags"),
        ), patch.object(
            self.conn,
            "describe_elasticsearch_domain",
            return_value={"DomainStatus": DOMAIN_RET},
        ):
            result = boto3_elasticsearch.remove_tags(
                tag_keys=["foo", "bar"], domain_name="testdomain", **CONN_PARAMETERS
            )
            self.assertFalse(result["result"])
            self.assertEqual(
                result.get("error", ""), ERROR_MESSAGE.format(101, "remove_tags")
            )

    def test_list_tags_positive(self):
        """
        Test that when listing tags is successful,
        the .list_tags method returns a dict with key 'tags'.
        Also test that the tags returned are manipulated properly (i.e. transformed
        into a dict with tags).
        """
        with patch.object(
            self.conn,
            "describe_elasticsearch_domain",
            return_value={"DomainStatus": DOMAIN_RET},
        ), patch.object(
            self.conn,
            "list_tags",
            return_value={"TagList": [{"Key": "foo", "Value": "bar"}]},
        ):
            result = boto3_elasticsearch.list_tags(
                domain_name="testdomain", **CONN_PARAMETERS
            )
            self.assertEqual(result, {"result": True, "response": {"foo": "bar"}})

    def test_list_tags_error(self):
        """
        Test that when listing tags causes boto3 to return an error,
        the .list_tags method returns the error.
        """
        with patch.object(
            self.conn, "list_tags", side_effect=ClientError(ERROR_CONTENT, "list_tags")
        ), patch.object(
            self.conn,
            "describe_elasticsearch_domain",
            return_value={"DomainStatus": DOMAIN_RET},
        ):
            result = boto3_elasticsearch.list_tags(
                domain_name="testdomain", **CONN_PARAMETERS
            )
            self.assertFalse(result["result"])
            self.assertEqual(
                result.get("error", ""), ERROR_MESSAGE.format(101, "list_tags")
            )

    def test_cancel_elasticsearch_service_software_update_positive(self):
        """
        Test that when calling cancel_elasticsearch_service_software_update and
        it is successful, it returns {'result': True}.
        """
        retval = {
            "ServiceSoftwareOptions": {
                "CurrentVersion": "string",
                "NewVersion": "string",
                "UpdateAvailable": True,
                "Cancellable": True,
                "UpdateStatus": "ELIGIBLE",
                "Description": "string",
                "AutomatedUpdateDate": datetime.datetime(2015, 1, 1),
            }
        }
        with patch.object(
            self.conn,
            "cancel_elasticsearch_service_software_update",
            return_value=retval,
        ):
            result = boto3_elasticsearch.cancel_elasticsearch_service_software_update(
                domain_name="testdomain", **CONN_PARAMETERS
            )
            self.assertEqual(result, {"result": True})

    def test_cancel_elasticsearch_service_software_update_error(self):
        """
        Test that when calling cancel_elasticsearch_service_software_update and
        boto3 returns an error, it returns {'result': False, 'error': 'the error'}.
        """
        with patch.object(
            self.conn,
            "cancel_elasticsearch_service_software_update",
            side_effect=ClientError(
                ERROR_CONTENT, "cancel_elasticsearch_service_software_update"
            ),
        ):
            result = boto3_elasticsearch.cancel_elasticsearch_service_software_update(
                domain_name="testdomain", **CONN_PARAMETERS
            )
            self.assertFalse(result["result"])
            self.assertEqual(
                result.get("error", ""),
                ERROR_MESSAGE.format(
                    101, "cancel_elasticsearch_service_software_update"
                ),
            )

    def test_delete_elasticsearch_service_role_positive(self):
        """
        Test that when calling delete_elasticsearch_service_role and
        it is successful, it returns {'result': True}.
        """
        with patch.object(
            self.conn, "delete_elasticsearch_service_role", return_value=None
        ):
            result = boto3_elasticsearch.delete_elasticsearch_service_role(
                **CONN_PARAMETERS
            )
            self.assertEqual(result, {"result": True})

    def test_delete_elasticsearch_service_role_error(self):
        """
        Test that when calling delete_elasticsearch_service_role and boto3 returns
        an error, it returns {'result': False, 'error': 'the error'}.
        """
        with patch.object(
            self.conn,
            "delete_elasticsearch_service_role",
            side_effect=ClientError(ERROR_CONTENT, "delete_elasticsearch_service_role"),
        ):
            result = boto3_elasticsearch.delete_elasticsearch_service_role(
                **CONN_PARAMETERS
            )
            self.assertFalse(result["result"])
            self.assertEqual(
                result.get("error", ""),
                ERROR_MESSAGE.format(101, "delete_elasticsearch_service_role"),
            )

    def test_describe_elasticsearch_domain_config_positive(self):
        """
        Test that when calling describe_elasticsearch_domain_config and
        it is successful, it returns {'result': True}.
        """
        with patch.object(
            self.conn,
            "describe_elasticsearch_domain_config",
            return_value={"DomainConfig": DOMAIN_RET},
        ):
            self.assertEqual(
                boto3_elasticsearch.describe_elasticsearch_domain_config(
                    "testdomain", **CONN_PARAMETERS
                ),
                {"result": True, "response": DOMAIN_RET},
            )

    def test_describe_elasticsearch_domain_config_error(self):
        """
        Test that when calling describe_elasticsearch_domain_config and boto3 returns
        an error, it returns {'result': False, 'error': 'the error'}.
        """
        with patch.object(
            self.conn,
            "describe_elasticsearch_domain_config",
            side_effect=ClientError(
                ERROR_CONTENT, "describe_elasticsearch_domain_config"
            ),
        ):
            result = boto3_elasticsearch.describe_elasticsearch_domain_config(
                domain_name="testdomain", **CONN_PARAMETERS
            )
            self.assertFalse(result["result"])
            self.assertEqual(
                result.get("error", ""),
                ERROR_MESSAGE.format(101, "describe_elasticsearch_domain_config"),
            )

    def test_describe_elasticsearch_domains_positive(self):
        """
        Test that when calling describe_elasticsearch_domains and it is successful,
        it returns {'result': True, 'response': some_data}.
        """
        with patch.object(
            self.conn,
            "describe_elasticsearch_domains",
            return_value={"DomainStatusList": [DOMAIN_RET]},
        ):
            self.assertEqual(
                boto3_elasticsearch.describe_elasticsearch_domains(
                    domain_names=["test_domain"], **CONN_PARAMETERS
                ),
                {"result": True, "response": [DOMAIN_RET]},
            )

    def test_describe_elasticsearch_domains_error(self):
        """
        Test that when calling describe_elasticsearch_domains and boto3 returns
        an error, it returns {'result': False, 'error': 'the error'}.
        """
        with patch.object(
            self.conn,
            "describe_elasticsearch_domains",
            side_effect=ClientError(ERROR_CONTENT, "describe_elasticsearch_domains"),
        ):
            result = boto3_elasticsearch.describe_elasticsearch_domains(
                domain_names=["testdomain"], **CONN_PARAMETERS
            )
            self.assertFalse(result["result"])
            self.assertEqual(
                result.get("error", ""),
                ERROR_MESSAGE.format(101, "describe_elasticsearch_domains"),
            )

    def test_describe_elasticsearch_instance_type_limits_positive(self):
        """
        Test that when calling describe_elasticsearch_instance_type_limits and
        it succeeds, it returns {'result': True, 'response' some_value}.
        """
        ret_val = {
            "LimitsByRole": {
                "string": {
                    "StorageTypes": [
                        {
                            "StorageTypeName": "string",
                            "StorageSubTypeName": "string",
                            "StorageTypeLimits": [
                                {"LimitName": "string", "LimitValues": ["string"]}
                            ],
                        }
                    ],
                    "InstanceLimits": {
                        "InstanceCountLimits": {
                            "MinimumInstanceCount": 123,
                            "MaximumInstanceCount": 123,
                        }
                    },
                    "AdditionalLimits": [
                        {"LimitName": "string", "LimitValues": ["string"]}
                    ],
                }
            }
        }
        with patch.object(
            self.conn,
            "describe_elasticsearch_instance_type_limits",
            return_value=ret_val,
        ):
            self.assertEqual(
                boto3_elasticsearch.describe_elasticsearch_instance_type_limits(
                    domain_name="testdomain",
                    instance_type="foo",
                    elasticsearch_version="1.0",
                    **CONN_PARAMETERS
                ),
                {"result": True, "response": ret_val["LimitsByRole"]},
            )

    def test_describe_elasticsearch_instance_type_limits_error(self):
        """
        Test that when calling describe_elasticsearch_instance_type_limits and boto3 returns
        an error, it returns {'result': False, 'error': 'the error'}.
        """
        with patch.object(
            self.conn,
            "describe_elasticsearch_instance_type_limits",
            side_effect=ClientError(
                ERROR_CONTENT, "describe_elasticsearch_instance_type_limits"
            ),
        ):
            result = boto3_elasticsearch.describe_elasticsearch_instance_type_limits(
                domain_name="testdomain",
                instance_type="foo",
                elasticsearch_version="1.0",
                **CONN_PARAMETERS
            )
            self.assertFalse(result["result"])
            self.assertEqual(
                result.get("error", ""),
                ERROR_MESSAGE.format(
                    101, "describe_elasticsearch_instance_type_limits"
                ),
            )

    def test_describe_reserved_elasticsearch_instance_offerings_positive(self):
        """
        Test that when calling describe_reserved_elasticsearch_instance_offerings
        and it succeeds, it returns {'result': True, 'response': some_value}.
        """
        ret_val = {
            "NextToken": "string",
            "ReservedElasticsearchInstanceOfferings": [
                {
                    "ReservedElasticsearchInstanceOfferingId": "string",
                    "ElasticsearchInstanceType": "t2.medium.elasticsearch",
                    "Duration": 123,
                    "FixedPrice": 123.0,
                    "UsagePrice": 123.0,
                    "CurrencyCode": "string",
                    "PaymentOption": "NO_UPFRONT",
                    "RecurringCharges": [
                        {
                            "RecurringChargeAmount": 123.0,
                            "RecurringChargeFrequency": "string",
                        }
                    ],
                }
            ],
        }
        with patch.object(self.paginator, "paginate", return_value=[ret_val]):
            self.assertEqual(
                boto3_elasticsearch.describe_reserved_elasticsearch_instance_offerings(
                    reserved_elasticsearch_instance_offering_id="foo", **CONN_PARAMETERS
                ),
                {
                    "result": True,
                    "response": ret_val["ReservedElasticsearchInstanceOfferings"],
                },
            )

    def test_describe_reserved_elasticsearch_instance_offerings_error(self):
        """
        Test that when calling describe_reserved_elasticsearch_instance_offerings
        and boto3 returns an error, it returns {'result': False, 'error': 'the error'}.
        """
        with patch.object(
            self.paginator,
            "paginate",
            side_effect=ClientError(
                ERROR_CONTENT, "describe_reserved_elasticsearch_instance_offerings"
            ),
        ):
            result = (
                boto3_elasticsearch.describe_reserved_elasticsearch_instance_offerings(
                    reserved_elasticsearch_instance_offering_id="foo", **CONN_PARAMETERS
                )
            )
            self.assertFalse(result["result"])
            self.assertEqual(
                result.get("error", ""),
                ERROR_MESSAGE.format(
                    101, "describe_reserved_elasticsearch_instance_offerings"
                ),
            )

    def test_describe_reserved_elasticsearch_instances_positive(self):
        """
        Test that when calling describe_reserved_elasticsearch_instances and it
        succeeds, it returns {'result': True, 'response': some_value}.
        """
        ret_val = {
            "NextToken": "string",
            "ReservedElasticsearchInstances": [
                {
                    "ReservationName": "string",
                    "ReservedElasticsearchInstanceId": "string",
                    "ReservedElasticsearchInstanceOfferingId": "string",
                    "ElasticsearchInstanceType": "t2.medium.elasticsearch",
                    "StartTime": datetime.datetime(2015, 1, 1),
                    "Duration": 123,
                    "FixedPrice": 123.0,
                    "UsagePrice": 123.0,
                    "CurrencyCode": "string",
                    "ElasticsearchInstanceCount": 123,
                    "State": "string",
                    "PaymentOption": "ALL_UPFRONT",
                    "RecurringCharges": [
                        {
                            "RecurringChargeAmount": 123.0,
                            "RecurringChargeFrequency": "string",
                        },
                    ],
                },
            ],
        }
        with patch.object(self.paginator, "paginate", return_value=[ret_val]):
            self.assertEqual(
                boto3_elasticsearch.describe_reserved_elasticsearch_instances(
                    reserved_elasticsearch_instance_id="foo", **CONN_PARAMETERS
                ),
                {"result": True, "response": ret_val["ReservedElasticsearchInstances"]},
            )

    def test_describe_reserved_elasticsearch_instances_error(self):
        """
        Test that when calling describe_reserved_elasticsearch_instances and boto3
        returns an error, it returns {'result': False, 'error': 'the error'}.
        """
        with patch.object(
            self.paginator,
            "paginate",
            side_effect=ClientError(
                ERROR_CONTENT, "describe_reserved_elasticsearch_instances"
            ),
        ):
            result = boto3_elasticsearch.describe_reserved_elasticsearch_instances(
                reserved_elasticsearch_instance_id="foo", **CONN_PARAMETERS
            )
            self.assertFalse(result["result"])
            self.assertEqual(
                result.get("error", ""),
                ERROR_MESSAGE.format(101, "describe_reserved_elasticsearch_instances"),
            )

    def test_get_compatible_elasticsearch_versions_positive(self):
        """
        Test that when calling get_compatible_elasticsearch_versions and it
        succeeds, it returns {'result': True, 'response': some_value}.
        """
        ret_val = {
            "CompatibleElasticsearchVersions": [
                {"SourceVersion": "string", "TargetVersions": ["string"]}
            ]
        }
        with patch.object(
            self.conn, "get_compatible_elasticsearch_versions", return_value=ret_val
        ):
            self.assertEqual(
                boto3_elasticsearch.get_compatible_elasticsearch_versions(
                    domain_name="testdomain", **CONN_PARAMETERS
                ),
                {
                    "result": True,
                    "response": ret_val["CompatibleElasticsearchVersions"],
                },
            )

    def test_get_compatible_elasticsearch_versions_error(self):
        """
        Test that when calling get_compatible_elasticsearch_versions and boto3
        returns an error, it returns {'result': False, 'error': 'the error'}.
        """
        with patch.object(
            self.conn,
            "get_compatible_elasticsearch_versions",
            side_effect=ClientError(
                ERROR_CONTENT, "get_compatible_elasticsearch_versions"
            ),
        ):
            result = boto3_elasticsearch.get_compatible_elasticsearch_versions(
                domain_name="testdomain", **CONN_PARAMETERS
            )
            self.assertFalse(result["result"])
            self.assertEqual(
                result.get("error", ""),
                ERROR_MESSAGE.format(101, "get_compatible_elasticsearch_versions"),
            )

    def test_get_upgrade_history_positive(self):
        """
        Test that when calling get_upgrade_history and it
        succeeds, it returns {'result': True, 'response': some_value}.
        """
        ret_val = {
            "UpgradeHistories": [
                {
                    "UpgradeName": "string",
                    "StartTimestamp": datetime.datetime(2015, 1, 1),
                    "UpgradeStatus": "IN_PROGRESS",
                    "StepsList": [
                        {
                            "UpgradeStep": "PRE_UPGRADE_CHECK",
                            "UpgradeStepStatus": "IN_PROGRESS",
                            "Issues": ["string"],
                            "ProgressPercent": 123.0,
                        }
                    ],
                }
            ],
            "NextToken": "string",
        }
        with patch.object(self.paginator, "paginate", return_value=[ret_val]):
            self.assertEqual(
                boto3_elasticsearch.get_upgrade_history(
                    domain_name="testdomain", **CONN_PARAMETERS
                ),
                {"result": True, "response": ret_val["UpgradeHistories"]},
            )

    def test_get_upgrade_history_error(self):
        """
        Test that when calling get_upgrade_history and boto3
        returns an error, it returns {'result': False, 'error': 'the error'}.
        """
        with patch.object(
            self.paginator,
            "paginate",
            side_effect=ClientError(ERROR_CONTENT, "get_upgrade_history"),
        ):
            result = boto3_elasticsearch.get_upgrade_history(
                domain_name="testdomain", **CONN_PARAMETERS
            )
            self.assertFalse(result["result"])
            self.assertEqual(
                result.get("error", ""),
                ERROR_MESSAGE.format(101, "get_upgrade_history"),
            )

    def test_get_upgrade_status_positive(self):
        """
        Test that when calling get_upgrade_status and it
        succeeds, it returns {'result': True, 'response': some_value}.
        """
        ret_val = {
            "UpgradeStep": "PRE_UPGRADE_CHECK",
            "StepStatus": "IN_PROGRESS",
            "UpgradeName": "string",
            "ResponseMetadata": None,
        }
        with patch.object(self.conn, "get_upgrade_status", return_value=ret_val):
            self.assertEqual(
                boto3_elasticsearch.get_upgrade_status(
                    domain_name="testdomain", **CONN_PARAMETERS
                ),
                {"result": True, "response": ret_val},
            )

    def test_get_upgrade_status_error(self):
        """
        Test that when calling get_upgrade_status and boto3
        returns an error, it returns {'result': False, 'error': 'the error'}.
        """
        with patch.object(
            self.conn,
            "get_upgrade_status",
            side_effect=ClientError(ERROR_CONTENT, "get_upgrade_status"),
        ):
            result = boto3_elasticsearch.get_upgrade_status(
                domain_name="testdomain", **CONN_PARAMETERS
            )
            self.assertFalse(result["result"])
            self.assertEqual(
                result.get("error", ""), ERROR_MESSAGE.format(101, "get_upgrade_status")
            )

    def test_list_domain_names_positive(self):
        """
        Test that when calling list_domain_names and it
        succeeds, it returns {'result': True, 'response': some_value}.
        """
        ret_val = {"DomainNames": [{"DomainName": "string"}]}
        with patch.object(self.conn, "list_domain_names", return_value=ret_val):
            self.assertEqual(
                boto3_elasticsearch.list_domain_names(**CONN_PARAMETERS),
                {
                    "result": True,
                    "response": [item["DomainName"] for item in ret_val["DomainNames"]],
                },
            )

    def test_list_domain_names_error(self):
        """
        Test that when calling list_domain_names and boto3
        returns an error, it returns {'result': False, 'error': 'the error'}.
        """
        with patch.object(
            self.conn,
            "list_domain_names",
            side_effect=ClientError(ERROR_CONTENT, "list_domain_names"),
        ):
            result = boto3_elasticsearch.list_domain_names(**CONN_PARAMETERS)
            self.assertFalse(result["result"])
            self.assertEqual(
                result.get("error", ""), ERROR_MESSAGE.format(101, "list_domain_names")
            )

    def test_list_elasticsearch_instance_types_positive(self):
        """
        Test that when calling list_elasticsearch_instance_types and it
        succeeds, it returns {'result': True, 'response': some_value}.
        """
        ret_val = {
            "ElasticsearchInstanceTypes": [
                "m3.medium.elasticsearch",
                "m3.large.elasticsearch",
                "m3.xlarge.elasticsearch",
                "m3.2xlarge.elasticsearch",
                "m4.large.elasticsearch",
                "m4.xlarge.elasticsearch",
                "m4.2xlarge.elasticsearch",
                "m4.4xlarge.elasticsearch",
                "m4.10xlarge.elasticsearch",
                "t2.micro.elasticsearch",
                "t2.small.elasticsearch",
                "t2.medium.elasticsearch",
                "r3.large.elasticsearch",
                "r3.xlarge.elasticsearch",
                "r3.2xlarge.elasticsearch",
                "r3.4xlarge.elasticsearch",
                "r3.8xlarge.elasticsearch",
                "i2.xlarge.elasticsearch",
                "i2.2xlarge.elasticsearch",
                "d2.xlarge.elasticsearch",
                "d2.2xlarge.elasticsearch",
                "d2.4xlarge.elasticsearch",
                "d2.8xlarge.elasticsearch",
                "c4.large.elasticsearch",
                "c4.xlarge.elasticsearch",
                "c4.2xlarge.elasticsearch",
                "c4.4xlarge.elasticsearch",
                "c4.8xlarge.elasticsearch",
                "r4.large.elasticsearch",
                "r4.xlarge.elasticsearch",
                "r4.2xlarge.elasticsearch",
                "r4.4xlarge.elasticsearch",
                "r4.8xlarge.elasticsearch",
                "r4.16xlarge.elasticsearch",
                "i3.large.elasticsearch",
                "i3.xlarge.elasticsearch",
                "i3.2xlarge.elasticsearch",
                "i3.4xlarge.elasticsearch",
                "i3.8xlarge.elasticsearch",
                "i3.16xlarge.elasticsearch",
            ],
            "NextToken": "string",
        }
        with patch.object(self.paginator, "paginate", return_value=[ret_val]):
            self.assertEqual(
                boto3_elasticsearch.list_elasticsearch_instance_types(
                    elasticsearch_version="1.0", **CONN_PARAMETERS
                ),
                {"result": True, "response": ret_val["ElasticsearchInstanceTypes"]},
            )

    def test_list_elasticsearch_instance_types_error(self):
        """
        Test that when calling list_elasticsearch_instance_types and boto3
        returns an error, it returns {'result': False, 'error': 'the error'}.
        """
        with patch.object(
            self.paginator,
            "paginate",
            side_effect=ClientError(ERROR_CONTENT, "list_elasticsearch_instance_types"),
        ):
            result = boto3_elasticsearch.list_elasticsearch_instance_types(
                elasticsearch_version="1.0", **CONN_PARAMETERS
            )
            self.assertFalse(result["result"])
            self.assertEqual(
                result.get("error", ""),
                ERROR_MESSAGE.format(101, "list_elasticsearch_instance_types"),
            )

    def test_list_elasticsearch_versions_positive(self):
        """
        Test that when calling list_elasticsearch_versions and it
        succeeds, it returns {'result': True, 'response': some_value}.
        """
        ret_val = {"ElasticsearchVersions": ["string"], "NextToken": "string"}
        with patch.object(self.paginator, "paginate", return_value=[ret_val]):
            self.assertEqual(
                boto3_elasticsearch.list_elasticsearch_versions(**CONN_PARAMETERS),
                {"result": True, "response": ret_val["ElasticsearchVersions"]},
            )

    def test_list_elasticsearch_versions_error(self):
        """
        Test that when calling list_elasticsearch_versions and boto3
        returns an error, it returns {'result': False, 'error': 'the error'}.
        """
        with patch.object(
            self.paginator,
            "paginate",
            side_effect=ClientError(ERROR_CONTENT, "list_elasticsearch_versions"),
        ):
            result = boto3_elasticsearch.list_elasticsearch_versions(**CONN_PARAMETERS)
            self.assertFalse(result["result"])
            self.assertEqual(
                result.get("error", ""),
                ERROR_MESSAGE.format(101, "list_elasticsearch_versions"),
            )

    def test_purchase_reserved_elasticsearch_instance_offering_positive(self):
        """
        Test that when calling purchase_reserved_elasticsearch_instance_offering and it
        succeeds, it returns {'result': True, 'response': some_value}.
        """
        ret_val = {
            "ReservedElasticsearchInstanceId": "string",
            "ReservationName": "string",
        }
        with patch.object(
            self.conn,
            "purchase_reserved_elasticsearch_instance_offering",
            return_value=ret_val,
        ):
            self.assertEqual(
                boto3_elasticsearch.purchase_reserved_elasticsearch_instance_offering(
                    reserved_elasticsearch_instance_offering_id="foo",
                    reservation_name="bar",
                    **CONN_PARAMETERS
                ),
                {"result": True, "response": ret_val},
            )

    def test_purchase_reserved_elasticsearch_instance_offering_error(self):
        """
        Test that when calling purchase_reserved_elasticsearch_instance_offering and boto3
        returns an error, it returns {'result': False, 'error': 'the error'}.
        """
        with patch.object(
            self.conn,
            "purchase_reserved_elasticsearch_instance_offering",
            side_effect=ClientError(
                ERROR_CONTENT, "purchase_reserved_elasticsearch_instance_offering"
            ),
        ):
            result = (
                boto3_elasticsearch.purchase_reserved_elasticsearch_instance_offering(
                    reserved_elasticsearch_instance_offering_id="foo",
                    reservation_name="bar",
                    **CONN_PARAMETERS
                )
            )
            self.assertFalse(result["result"])
            self.assertEqual(
                result.get("error", ""),
                ERROR_MESSAGE.format(
                    101, "purchase_reserved_elasticsearch_instance_offering"
                ),
            )

    def test_start_elasticsearch_service_software_update_positive(self):
        """
        Test that when calling start_elasticsearch_service_software_update and it
        succeeds, it returns {'result': True, 'response': some_value}.
        """
        ret_val = {
            "ServiceSoftwareOptions": {
                "CurrentVersion": "string",
                "NewVersion": "string",
                "UpdateAvailable": True,
                "Cancellable": True,
                "UpdateStatus": "PENDING_UPDATE",
                "Description": "string",
                "AutomatedUpdateDate": datetime.datetime(2015, 1, 1),
            }
        }
        with patch.object(
            self.conn,
            "start_elasticsearch_service_software_update",
            return_value=ret_val,
        ):
            self.assertEqual(
                boto3_elasticsearch.start_elasticsearch_service_software_update(
                    domain_name="testdomain", **CONN_PARAMETERS
                ),
                {"result": True, "response": ret_val["ServiceSoftwareOptions"]},
            )

    def test_start_elasticsearch_service_software_update_error(self):
        """
        Test that when calling start_elasticsearch_service_software_update and boto3
        returns an error, it returns {'result': False, 'error': 'the error'}.
        """
        with patch.object(
            self.conn,
            "start_elasticsearch_service_software_update",
            side_effect=ClientError(
                ERROR_CONTENT, "start_elasticsearch_service_software_update"
            ),
        ):
            result = boto3_elasticsearch.start_elasticsearch_service_software_update(
                domain_name="testdomain", **CONN_PARAMETERS
            )
            self.assertFalse(result["result"])
            self.assertEqual(
                result.get("error", ""),
                ERROR_MESSAGE.format(
                    101, "start_elasticsearch_service_software_update"
                ),
            )

    def test_upgrade_elasticsearch_domain_positive(self):
        """
        Test that when calling upgrade_elasticsearch_domain and it
        succeeds, it returns {'result': True, 'response': some_value}.
        """
        ret_val = {
            "DomainName": "string",
            "TargetVersion": "string",
            "PerformCheckOnly": True,
        }
        with patch.object(
            self.conn, "upgrade_elasticsearch_domain", return_value=ret_val
        ):
            self.assertEqual(
                boto3_elasticsearch.upgrade_elasticsearch_domain(
                    domain_name="testdomain", target_version="1.1", **CONN_PARAMETERS
                ),
                {"result": True, "response": ret_val},
            )

    def test_upgrade_elasticsearch_domain_error(self):
        """
        Test that when calling upgrade_elasticsearch_domain and boto3
        returns an error, it returns {'result': False, 'error': 'the error'}.
        """
        with patch.object(
            self.conn,
            "upgrade_elasticsearch_domain",
            side_effect=ClientError(ERROR_CONTENT, "upgrade_elasticsearch_domain"),
        ):
            result = boto3_elasticsearch.upgrade_elasticsearch_domain(
                domain_name="testdomain", target_version="1.1", **CONN_PARAMETERS
            )
            self.assertFalse(result["result"])
            self.assertEqual(
                result.get("error", ""),
                ERROR_MESSAGE.format(101, "upgrade_elasticsearch_domain"),
            )



"""
Management of OpenStack Neutron Security Group Rules
====================================================

.. versionadded:: 2018.3.0

:depends: shade
:configuration: see :py:mod:`salt.modules.neutronng` for setup instructions

Example States

.. code-block:: yaml

    create security group rule:
      neutron_secgroup_rule.present:
        - name: security_group1
        - project_name: Project1
        - protocol: icmp

    delete security group:
      neutron_secgroup_rule.absent:
        - name_or_id: security_group1

    create security group with optional params:
      neutron_secgroup_rule.present:
        - name: security_group1
        - description: "Very Secure Security Group"
        - project_id: 1dcac318a83b4610b7a7f7ba01465548
"""


__virtualname__ = "neutron_secgroup_rule"


def __virtual__():
    if "neutronng.list_subnets" in __salt__:
        return __virtualname__
    return (
        False,
        "The neutronng execution module failed to load: shade python module is not available",
    )


def _rule_compare(rule1, rule2):
    """
    Compare the common keys between security group rules against eachother
    """

    commonkeys = set(rule1.keys()).intersection(rule2.keys())
    for key in commonkeys:
        if rule1[key] != rule2[key]:
            return False
    return True


def present(name, auth=None, **kwargs):
    """
    Ensure a security group rule exists

    defaults: port_range_min=None, port_range_max=None, protocol=None,
              remote_ip_prefix=None, remote_group_id=None, direction='ingress',
              ethertype='IPv4', project_id=None

    name
        Name of the security group to associate with this rule

    project_name
        Name of the project associated with the security group

    protocol
        The protocol that is matched by the security group rule.
        Valid values are None, tcp, udp, and icmp.

    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    kwargs = __utils__["args.clean_kwargs"](**kwargs)

    __salt__["neutronng.setup_clouds"](auth)

    if "project_name" in kwargs:
        kwargs["project_id"] = kwargs["project_name"]
        del kwargs["project_name"]

    project = __salt__["keystoneng.project_get"](name=kwargs["project_id"])

    if project is None:
        ret["result"] = False
        ret["comment"] = "Project does not exist"
        return ret

    secgroup = __salt__["neutronng.security_group_get"](
        name=name, filters={"tenant_id": project.id}
    )

    if secgroup is None:
        ret["result"] = False
        ret["changes"] = ({},)
        ret["comment"] = "Security Group does not exist {}".format(name)
        return ret

    # we have to search through all secgroup rules for a possible match
    rule_exists = None
    for rule in secgroup["security_group_rules"]:
        if _rule_compare(rule, kwargs) is True:
            rule_exists = True

    if rule_exists is None:
        if __opts__["test"] is True:
            ret["result"] = None
            ret["changes"] = kwargs
            ret["comment"] = "Security Group rule will be created."
            return ret

        # The variable differences are a little clumsy right now
        kwargs["secgroup_name_or_id"] = secgroup

        new_rule = __salt__["neutronng.security_group_rule_create"](**kwargs)
        ret["changes"] = new_rule
        ret["comment"] = "Created security group rule"
        return ret

    return ret


def absent(name, auth=None, **kwargs):
    """
    Ensure a security group rule does not exist

    name
        name or id of the security group rule to delete

    rule_id
        uuid of the rule to delete

    project_id
        id of project to delete rule from
    """
    rule_id = kwargs["rule_id"]
    ret = {"name": rule_id, "changes": {}, "result": True, "comment": ""}

    __salt__["neutronng.setup_clouds"](auth)

    secgroup = __salt__["neutronng.security_group_get"](
        name=name, filters={"tenant_id": kwargs["project_id"]}
    )

    # no need to delete a rule if the security group doesn't exist
    if secgroup is None:
        ret["comment"] = "security group does not exist"
        return ret

    # This should probably be done with compare on fields instead of
    # rule_id in the future
    rule_exists = None
    for rule in secgroup["security_group_rules"]:
        if _rule_compare(rule, {"id": rule_id}) is True:
            rule_exists = True

    if rule_exists:
        if __opts__["test"]:
            ret["result"] = None
            ret["changes"] = {"id": kwargs["rule_id"]}
            ret["comment"] = "Security group rule will be deleted."
            return ret

        __salt__["neutronng.security_group_rule_delete"](rule_id=rule_id)
        ret["changes"]["id"] = rule_id
        ret["comment"] = "Deleted security group rule"

    return ret



import logging
import socket

import attr

import salt.auth
import salt.ext.tornado.escape
import salt.ext.tornado.web
from salt.ext.tornado import netutil
from salt.ext.tornado.httpclient import AsyncHTTPClient, HTTPError
from salt.ext.tornado.httpserver import HTTPServer
from salt.ext.tornado.ioloop import TimeoutError as IOLoopTimeoutError
from salt.netapi.rest_tornado import saltnado

log = logging.getLogger(__name__)


@attr.s
class TestsHttpClient:
    address = attr.ib()
    io_loop = attr.ib(repr=False)
    headers = attr.ib(default=None)
    client = attr.ib(init=False, repr=False)

    @client.default
    def _client_default(self):
        return AsyncHTTPClient(self.io_loop)

    async def fetch(self, path, **kwargs):
        if "headers" not in kwargs and self.headers:
            kwargs["headers"] = self.headers.copy()
        try:
            response = await self.client.fetch(
                "{}{}".format(self.address, path), **kwargs
            )
            return self._decode_body(response)
        except HTTPError as exc:
            exc.response = self._decode_body(exc.response)
            raise

    def _decode_body(self, response):
        if response is None:
            return response
        if response.body:
            # Decode it
            if response.headers.get("Content-Type") == "application/json":
                response._body = response.body.decode("utf-8")
            else:
                response._body = salt.ext.tornado.escape.native_str(response.body)
        return response


@attr.s
class TestsTornadoHttpServer:
    io_loop = attr.ib(repr=False)
    app = attr.ib()
    port = attr.ib(repr=False)
    protocol = attr.ib(default="http", repr=False)
    http_server_options = attr.ib(default=attr.Factory(dict))
    sock = attr.ib(init=False, repr=False)
    address = attr.ib(init=False)
    server = attr.ib(init=False)
    client_headers = attr.ib(default=None)
    client = attr.ib(init=False, repr=False)

    @sock.default
    def _sock_default(self):
        return netutil.bind_sockets(
            self.port, "127.0.0.1", family=socket.AF_INET, reuse_port=False
        )[0]

    @port.default
    def _port_default(self):
        return self.sock.getsockname()[1]

    @address.default
    def _address_default(self):
        return "{}://127.0.0.1:{}".format(self.protocol, self.port)

    @server.default
    def _server_default(self):
        server = HTTPServer(self.app, io_loop=self.io_loop, **self.http_server_options)
        server.add_sockets([self.sock])
        return server

    @client.default
    def _client_default(self):
        return TestsHttpClient(
            address=self.address, io_loop=self.io_loop, headers=self.client_headers
        )

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.server.stop()
        try:
            self.io_loop.run_sync(self.server.close_all_connections, timeout=10)
        except IOLoopTimeoutError:
            pass
        self.client.client.close()


def load_auth(client_config):
    return salt.auth.LoadAuth(client_config)


def auth_token(load_auth, auth_creds):
    """
    Mint and return a valid token for auth_creds
    """
    return load_auth.mk_token(auth_creds)


def build_tornado_app(
    urls, load_auth, client_config, minion_config, setup_event_listener=False
):
    application = salt.ext.tornado.web.Application(urls, debug=True)

    application.auth = load_auth
    application.opts = client_config
    application.mod_opts = minion_config
    if setup_event_listener:
        application.event_listener = saltnado.EventListener(
            minion_config, client_config
        )
    return application


def content_type_map():
    return {
        "json": "application/json",
        "json-utf8": "application/json; charset=utf-8",
        "yaml": "application/x-yaml",
        "text": "text/plain",
        "form": "application/x-www-form-urlencoded",
        "xml": "application/xml",
        "real-accept-header-json": "application/json, text/javascript, */*; q=0.01",
        "real-accept-header-yaml": "application/x-yaml, text/yaml, */*; q=0.01",
    }



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import os
import traceback

import pytest
from saltfactories.utils.tempfiles import temp_file

import salt.config
import salt.utils.yaml
from salt.output import display_output
from tests.support.case import ShellCase
from tests.support.mixins import RUNTIME_VARS


class OutputReturnTest(ShellCase):
    """
    Integration tests to ensure outputters return their expected format.
    Tests against situations where the loader might not be returning the
    right outputter even though it was explicitly requested.
    """

    @pytest.mark.slow_test
    def test_output_json(self):
        """
        Tests the return of json-formatted data
        """
        ret = self.run_call("test.ping --out=json")
        self.assertIn("{", ret)
        self.assertIn('"local": true', "".join(ret))
        self.assertIn("}", "".join(ret))

    @pytest.mark.slow_test
    def test_output_nested(self):
        """
        Tests the return of nested-formatted data
        """
        expected = ["local:", "    True"]
        ret = self.run_call("test.ping --out=nested")
        self.assertEqual(ret, expected)

    @pytest.mark.slow_test
    def test_output_quiet(self):
        """
        Tests the return of an out=quiet query
        """
        expected = []
        ret = self.run_call("test.ping --out=quiet")
        self.assertEqual(ret, expected)

    @pytest.mark.slow_test
    def test_output_pprint(self):
        """
        Tests the return of pprint-formatted data
        """
        expected = ["{'local': True}"]
        ret = self.run_call("test.ping --out=pprint")
        self.assertEqual(ret, expected)

    @pytest.mark.slow_test
    def test_output_raw(self):
        """
        Tests the return of raw-formatted data
        """
        expected = ["{'local': True}"]
        ret = self.run_call("test.ping --out=raw")
        self.assertEqual(ret, expected)

    @pytest.mark.slow_test
    def test_output_txt(self):
        """
        Tests the return of txt-formatted data
        """
        expected = ["local: True"]
        ret = self.run_call("test.ping --out=txt")
        self.assertEqual(ret, expected)

    @pytest.mark.slow_test
    def test_output_yaml(self):
        """
        Tests the return of yaml-formatted data
        """
        expected = ["local: true"]
        ret = self.run_call("test.ping --out=yaml")
        self.assertEqual(ret, expected)

    @pytest.mark.slow_test
    def test_output_yaml_namespaced_dict_wrapper(self):
        """
        Tests the ability to dump a NamespacedDictWrapper instance, as used in
        magic dunders like __grains__ and __pillar__

        See https://github.com/saltstack/salt/issues/49269
        """
        dumped_yaml = "\n".join(self.run_call("grains.items --out=yaml"))
        loaded_yaml = salt.utils.yaml.safe_load(dumped_yaml)
        # We just want to check that the dumped YAML loades as a dict with a
        # single top-level key, we don't care about the real contents.
        assert isinstance(loaded_yaml, dict)
        assert list(loaded_yaml) == ["local"]

    def test_output_unicodebad(self):
        """
        Tests outputter reliability with utf8
        """
        opts = salt.config.minion_config(
            os.path.join(RUNTIME_VARS.TMP_CONF_DIR, "minion")
        )
        opts["output_file"] = os.path.join(RUNTIME_VARS.TMP, "outputtest")
        data = {"foo": {"result": False, "aaa": "azerzaeréééé", "comment": "ééééàààà"}}
        try:
            # this should not raises UnicodeEncodeError
            display_output(data, opts=opts)
        except Exception:  # pylint: disable=broad-except
            # display trace in error message for debugging on jenkins
            trace = traceback.format_exc()
            sentinel = object()
            old_max_diff = getattr(self, "maxDiff", sentinel)
            try:
                self.maxDiff = None
                self.assertEqual(trace, "")
            finally:
                if old_max_diff is sentinel:
                    delattr(self, "maxDiff")
                else:
                    self.maxDiff = old_max_diff

    @pytest.mark.slow_test
    def test_output_highstate(self):
        """
        Regression tests for the highstate outputter. Calls a basic state with various
        flags. Each comparison should be identical when successful.
        """
        simple_ping_sls = """
        simple-ping:
          module.run:
            - name: test.ping
        """
        with temp_file(
            "simple-ping.sls", simple_ping_sls, RUNTIME_VARS.TMP_BASEENV_STATE_TREE
        ):
            # Test basic highstate output. No frills.
            expected = [
                "minion:",
                "          ID: simple-ping",
                "    Function: module.run",
                "        Name: test.ping",
                "      Result: True",
                "     Comment: Module function test.ping executed",
                "     Changes:   ",
                "              ret:",
                "                  True",
                "Summary for minion",
                "Succeeded: 1 (changed=1)",
                "Failed:    0",
                "Total states run:     1",
            ]
            state_run = self.run_salt('"minion" state.sls simple-ping')

            for expected_item in expected:
                self.assertIn(expected_item, state_run)

            # Test highstate output while also passing --out=highstate.
            # This is a regression test for Issue #29796
            state_run = self.run_salt('"minion" state.sls simple-ping --out=highstate')

            for expected_item in expected:
                self.assertIn(expected_item, state_run)

            # Test highstate output when passing --static and running a state function.
            # See Issue #44556.
            state_run = self.run_salt('"minion" state.sls simple-ping --static')

            for expected_item in expected:
                self.assertIn(expected_item, state_run)

            # Test highstate output when passing --static and --out=highstate.
            # See Issue #44556.
            state_run = self.run_salt(
                '"minion" state.sls simple-ping --static --out=highstate'
            )

            for expected_item in expected:
                self.assertIn(expected_item, state_run)

    @pytest.mark.slow_test
    def test_output_highstate_falls_back_nested(self):
        """
        Tests outputter when passing --out=highstate with a non-state call. This should
        fall back to "nested" output.
        """
        expected = ["minion:", "    True"]
        ret = self.run_salt('"minion" test.ping --out=highstate')
        self.assertEqual(ret, expected)

    @pytest.mark.slow_test
    def test_static_simple(self):
        """
        Tests passing the --static option with a basic test.ping command. This
        should be the "nested" output.
        """
        expected = ["minion:", "    True"]
        ret = self.run_salt('"minion" test.ping --static')
        self.assertEqual(ret, expected)



"""
PyInstaller runtime hook to patch subprocess.Popen
"""
import logging
import subprocess

from salt.utils.pyinstaller.rthooks._overrides import PyinstallerPopen

log = logging.getLogger(__name__)
# Patch subprocess.Popen when running within a pyinstalled bundled package
subprocess.Popen = PyinstallerPopen

log.debug("Replaced 'subprocess.Popen' with 'PyinstallerTerminal'")



"""
Tests written specifically for the grains.append function.
"""

import logging
import time

import attr
import pytest

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.windows_whitelisted,
]


@attr.s(frozen=True, slots=True)
class AppendGrain:
    key = attr.ib(default="append-grains-test-key")
    value = attr.ib(default="my-grain-value")


@pytest.fixture(scope="module")
def append_grain_module(salt_call_cli, wait_for_pillar_refresh_complete):
    grain = AppendGrain()
    try:
        # Start off with an empty list
        start_time = time.time()
        ret = salt_call_cli.run("grains.setval", grain.key, val=[])
        assert ret.returncode == 0
        assert ret.data
        assert ret.data == {grain.key: []}

        # Let's wait for the pillar refresh, at which stage we know grains are also refreshed
        wait_for_pillar_refresh_complete(start_time)
        yield grain
    finally:
        start_time = time.time()
        ret = salt_call_cli.run("grains.delkey", grain.key, force=True)
        assert ret.returncode == 0
        assert ret.data

        # Let's wait for the pillar refresh, at which stage we know grains are also refreshed
        wait_for_pillar_refresh_complete(start_time)


@pytest.fixture
def append_grain(append_grain_module, salt_call_cli, wait_for_pillar_refresh_complete):
    try:
        yield append_grain_module
    finally:
        start_time = time.time()
        ret = salt_call_cli.run("grains.setval", append_grain_module.key, val=[])
        assert ret.returncode == 0
        assert ret.data
        assert ret.data == {append_grain_module.key: []}

        # Let's wait for the pillar refresh, at which stage we know grains are also refreshed
        wait_for_pillar_refresh_complete(start_time)


def test_grains_append(salt_call_cli, append_grain):
    """
    Tests the return of a simple grains.append call.
    """
    ret = salt_call_cli.run("grains.append", append_grain.key, append_grain.value)
    assert ret.returncode == 0
    assert ret.data
    assert ret.data == {append_grain.key: [append_grain.value]}


def test_grains_append_val_already_present(salt_call_cli, append_grain):
    """
    Tests the return of a grains.append call when the value is already
    present in the grains list.
    """
    msg = "The val {} was already in the list {}".format(
        append_grain.value, append_grain.key
    )

    # First, make sure the test grain is present
    ret = salt_call_cli.run("grains.append", append_grain.key, append_grain.value)
    assert ret.returncode == 0
    assert ret.data
    assert ret.data == {append_grain.key: [append_grain.value]}

    # Now try to append again
    ret = salt_call_cli.run("grains.append", append_grain.key, append_grain.value)
    assert ret.returncode == 0
    assert ret.data
    assert ret.data == msg


def test_grains_append_val_is_list(salt_call_cli, append_grain):
    """
    Tests the return of a grains.append call when val is passed in as a list.
    """
    second_grain = append_grain.value + "-2"
    ret = salt_call_cli.run(
        "grains.append", append_grain.key, val=[append_grain.value, second_grain]
    )
    assert ret.returncode == 0
    assert ret.data
    assert ret.data == {append_grain.key: [append_grain.value, second_grain]}


def test_grains_remove_add(
    salt_call_cli, append_grain, wait_for_pillar_refresh_complete
):
    second_grain = append_grain.value + "-2"
    ret = salt_call_cli.run("grains.get", append_grain.key)
    assert ret.returncode == 0
    assert ret.data == []

    # The range was previously set to 10. Honestly, I don't know why testing 2 iterations
    # would be any different than 10. Maybe because we're making salt work harder...
    # Anyway, setting at 3 since it sounds more reasonable.
    for _ in range(3):
        start_time = time.time()
        ret = salt_call_cli.run("grains.setval", append_grain.key, val=[])
        assert ret.returncode == 0
        assert ret.data
        assert ret.data == {append_grain.key: []}
        wait_for_pillar_refresh_complete(start_time)
        ret = salt_call_cli.run("grains.get", append_grain.key)
        assert ret.returncode == 0
        assert ret.data == []

        start_time = time.time()
        ret = salt_call_cli.run("grains.append", append_grain.key, append_grain.value)
        assert ret.returncode == 0
        assert ret.data
        assert ret.data == {append_grain.key: [append_grain.value]}
        wait_for_pillar_refresh_complete(start_time)
        ret = salt_call_cli.run("grains.get", append_grain.key)
        assert ret.returncode == 0
        assert ret.data == [append_grain.value]

        start_time = time.time()
        ret = salt_call_cli.run("grains.setval", append_grain.key, val=[])
        assert ret.returncode == 0
        assert ret.data
        assert ret.data == {append_grain.key: []}
        wait_for_pillar_refresh_complete(start_time)
        ret = salt_call_cli.run("grains.get", append_grain.key)
        assert ret.returncode == 0
        assert ret.data == []

        start_time = time.time()
        ret = salt_call_cli.run(
            "grains.append", append_grain.key, val=[append_grain.value, second_grain]
        )
        assert ret.returncode == 0
        assert ret.data
        assert ret.data == {append_grain.key: [append_grain.value, second_grain]}
        wait_for_pillar_refresh_complete(start_time)
        ret = salt_call_cli.run("grains.get", append_grain.key)
        assert ret.returncode == 0
        assert ret.data == [append_grain.value, second_grain]



"""
    :codeauthor: :email:`Jakub Sliva <jakub.sliva@ultimum.io>`
"""

import pytest

import salt.states.zabbix_template as zabbix_template
from tests.support.mock import MagicMock, patch


@pytest.fixture
def defined_obj():
    return {
        "macros": [{"macro": "{$CEPH_CLUSTER_NAME}", "value": "ceph"}],
        "host": "A Testing Template",
        "hosts": [{"hostid": "10112"}, {"hostid": "10113"}],
        "description": "Template for Ceph nodes",
        "groups": [{"groupid": "1"}],
    }


@pytest.fixture
def defined_c_list_subs():
    return {
        "applications": [{"name": "Ceph OSD"}],
        "graphs": [],
        "triggers": [],
        "items": [],
        "httpTests": [],
        "screens": [],
        "gitems": [],
        "discoveries": [],
    }


@pytest.fixture
def substitute_params_create(defined_obj, defined_c_list_subs):
    return [
        defined_obj,
        [],
        defined_c_list_subs["applications"],
        [],
        [],
        [],
        [],
        [],
        [],
        [],
        [],
    ]


@pytest.fixture
def existing_obj():
    return [
        {
            "available": "0",
            "tls_connect": "1",
            "maintenance_type": "0",
            "groups": [{"groupid": "1"}],
            "macros": [
                {
                    "macro": "{$CEPH_CLUSTER_NAME}",
                    "hostmacroid": "60",
                    "hostid": "10206",
                    "value": "ceph",
                }
            ],
            "hosts": [{"hostid": "10112"}, {"hostid": "10113"}],
            "status": "3",
            "description": "Template for Ceph nodes",
            "host": "A Testing Template",
            "disable_until": "0",
            "templateid": "10206",
            "name": "A Testing Template",
        }
    ]


@pytest.fixture
def substitute_params_exists(defined_obj, existing_obj):
    return [
        defined_obj,
        existing_obj[0],
        [],
        [],
        [],
        [],
        [],
        [],
        [],
        [],
    ]


@pytest.fixture
def existing_obj_diff():
    return [
        {
            "groups": [{"groupid": "1"}],
            "macros": [
                {
                    "macro": "{$CEPH_CLUSTER_NAME}",
                    "hostmacroid": "60",
                    "hostid": "10206",
                    "value": "ceph",
                }
            ],
            "hosts": [{"hostid": "10112"}, {"hostid": "10113"}],
            "status": "3",
            "templateid": "10206",
            "name": "A Testing Template",
        }
    ]


@pytest.fixture
def substitute_params_update(defined_obj, existing_obj_diff):
    return [
        defined_obj,
        existing_obj_diff[0],
        [],
        [],
        [],
        [],
        [],
        [],
        [],
        [],
    ]


@pytest.fixture
def diff_params():
    return {"old": {}, "new": {"macros": [], "templateid": "10206"}}


@pytest.fixture
def configure_loader_modules():
    return {zabbix_template: {}}


def test_present_create(substitute_params_create):
    """
    Test to ensure that named template is created
    """
    with patch("salt.states.zabbix_template.CHANGE_STACK", []):
        name = "A Testing Template"
        ret = {"name": name, "result": False, "comment": "", "changes": {}}

        def side_effect_run_query(*args):
            """
            Differentiate between __salt__ exec module function calls with different parameters.
            """
            if args[0] in ("template.get", "application.get"):
                return []
            elif args[0] == "template.create":
                return {"templateids": ["10206"]}
            elif args[0] == "application.create":
                return {"applicationids": ["701"]}

        with patch.dict(zabbix_template.__opts__, {"test": False}):
            with patch.dict(
                zabbix_template.__salt__,
                {
                    "zabbix.get_zabbix_id_mapper": MagicMock(
                        return_value={"template": "templateid"}
                    ),
                    "zabbix.substitute_params": MagicMock(
                        side_effect=substitute_params_create
                    ),
                    "zabbix.run_query": MagicMock(side_effect=side_effect_run_query),
                    "zabbix.compare_params": MagicMock(return_value={}),
                },
            ):
                ret["result"] = True
                ret["comment"] = 'Zabbix Template "{}" created.'.format(name)
                ret["changes"] = {
                    name: {
                        "old": 'Zabbix Template "{}" did not exist.'.format(name),
                        "new": (
                            'Zabbix Template "{}" created according definition.'.format(
                                name
                            )
                        ),
                    }
                }
                assert zabbix_template.present(name, {}) == ret


def test_present_exists(existing_obj, substitute_params_exists):
    """
    Test to ensure that named template is present and not changed
    """
    with patch("salt.states.zabbix_template.CHANGE_STACK", []):
        name = "A Testing Template"
        ret = {"name": name, "result": False, "comment": "", "changes": {}}

        def side_effect_run_query(*args):
            """
            Differentiate between __salt__ exec module function calls with different parameters.
            """
            if args[0] == "template.get":
                return existing_obj
            elif args[0] == "application.get":
                return ["non-empty"]

        with patch.dict(zabbix_template.__opts__, {"test": False}):
            with patch.dict(
                zabbix_template.__salt__,
                {
                    "zabbix.get_zabbix_id_mapper": MagicMock(
                        return_value={"template": "templateid"}
                    ),
                    "zabbix.substitute_params": MagicMock(
                        side_effect=substitute_params_exists
                    ),
                    "zabbix.run_query": MagicMock(side_effect=side_effect_run_query),
                    "zabbix.compare_params": MagicMock(
                        return_value={"new": {}, "old": {}}
                    ),
                },
            ):
                ret["result"] = True
                ret["comment"] = (
                    'Zabbix Template "{}" already exists and corresponds to a'
                    " definition.".format(name)
                )
                assert zabbix_template.present(name, {}) == ret


def test_present_update(diff_params, substitute_params_update):
    """
    Test to ensure that named template is present but must be updated
    """
    with patch("salt.states.zabbix_template.CHANGE_STACK", []):
        name = "A Testing Template"
        ret = {"name": name, "result": False, "comment": "", "changes": {}}

        def side_effect_run_query(*args):
            """
            Differentiate between __salt__ exec module function calls with different parameters.
            """
            if args[0] == "template.get":
                return ["length of result is 1 = template exists"]
            elif args[0] == "template.update":
                return diff_params

        with patch.dict(zabbix_template.__opts__, {"test": False}):
            with patch.dict(
                zabbix_template.__salt__,
                {
                    "zabbix.get_zabbix_id_mapper": MagicMock(
                        return_value={"template": "templateid"}
                    ),
                    "zabbix.substitute_params": MagicMock(
                        side_effect=substitute_params_update
                    ),
                    "zabbix.run_query": MagicMock(side_effect=side_effect_run_query),
                    "zabbix.compare_params": MagicMock(return_value=diff_params),
                },
            ):
                ret["result"] = True
                ret["comment"] = 'Zabbix Template "{}" updated.'.format(name)
                ret["changes"] = {
                    name: {
                        "old": 'Zabbix Template "{}" differed.'.format(name),
                        "new": (
                            'Zabbix Template "{}" updated according definition.'.format(
                                name
                            )
                        ),
                    }
                }
                assert zabbix_template.present(name, {}) == ret


def test_absent_test_mode():
    """
    Test to ensure that named template is absent in test mode
    """
    name = "A Testing Template"
    ret = {"name": name, "result": False, "comment": "", "changes": {}}
    with patch.dict(zabbix_template.__opts__, {"test": True}):
        with patch.dict(
            zabbix_template.__salt__,
            {"zabbix.get_object_id_by_params": MagicMock(return_value=11)},
        ):
            ret["result"] = True
            ret["comment"] = 'Zabbix Template "{}" would be deleted.'.format(name)
            ret["changes"] = {
                name: {
                    "old": 'Zabbix Template "{}" exists.'.format(name),
                    "new": 'Zabbix Template "{}" would be deleted.'.format(name),
                }
            }
            assert zabbix_template.absent(name) == ret


def test_absent():
    """
    Test to ensure that named template is absent
    """
    name = "A Testing Template"
    ret = {"name": name, "result": False, "comment": "", "changes": {}}
    with patch.dict(zabbix_template.__opts__, {"test": False}):
        with patch.dict(
            zabbix_template.__salt__,
            {"zabbix.get_object_id_by_params": MagicMock(return_value=False)},
        ):
            ret["result"] = True
            ret["comment"] = 'Zabbix Template "{}" does not exist.'.format(name)
            assert zabbix_template.absent(name) == ret

        with patch.dict(
            zabbix_template.__salt__,
            {"zabbix.get_object_id_by_params": MagicMock(return_value=11)},
        ):
            with patch.dict(
                zabbix_template.__salt__,
                {"zabbix.run_query": MagicMock(return_value=True)},
            ):
                ret["result"] = True
                ret["comment"] = 'Zabbix Template "{}" deleted.'.format(name)
                ret["changes"] = {
                    name: {
                        "old": 'Zabbix Template "{}" existed.'.format(name),
                        "new": 'Zabbix Template "{}" deleted.'.format(name),
                    }
                }
                assert zabbix_template.absent(name) == ret



import json

import pytest

from salt.defaults.exitcodes import EX_AGGREGATE

pytestmark = [
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
]


@pytest.fixture(scope="module")
def state_tree(base_env_state_tree_root_dir):
    top_file = """
    {%- from "map.jinja" import abc with context %}
    base:
      'localhost':
        - basic
      '127.0.0.1':
        - basic
    """
    map_file = """
    {%- set abc = "def" %}
    """
    state_file = """
    {%- from "map.jinja" import abc with context %}
    Ok with {{ abc }}:
      test.succeed_without_changes
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_state_tree_root_dir
    )
    map_tempfile = pytest.helpers.temp_file(
        "map.jinja", map_file, base_env_state_tree_root_dir
    )
    state_tempfile = pytest.helpers.temp_file(
        "test.sls", state_file, base_env_state_tree_root_dir
    )
    with top_tempfile, map_tempfile, state_tempfile:
        yield


@pytest.fixture(scope="module")
def state_tree_dir(base_env_state_tree_root_dir):
    """
    State tree with files to test salt-ssh
    when the map.jinja file is in another directory
    """
    top_file = """
    {%- from "test/map.jinja" import abc with context %}
    base:
      'localhost':
        - test
      '127.0.0.1':
        - test
    """
    map_file = """
    {%- set abc = "def" %}
    """
    state_file = """
    {%- from "test/map.jinja" import abc with context %}

    Ok with {{ abc }}:
      test.succeed_without_changes
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_state_tree_root_dir
    )
    map_tempfile = pytest.helpers.temp_file(
        "test/map.jinja", map_file, base_env_state_tree_root_dir
    )
    state_tempfile = pytest.helpers.temp_file(
        "test.sls", state_file, base_env_state_tree_root_dir
    )

    with top_tempfile, map_tempfile, state_tempfile:
        yield


@pytest.fixture(scope="class")
def state_tree_render_fail(base_env_state_tree_root_dir):
    top_file = """
    base:
      'localhost':
        - fail_render
      '127.0.0.1':
        - fail_render
    """
    state_file = r"""
    abc var is not defined {{ abc }}:
      test.nop
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_state_tree_root_dir
    )
    state_tempfile = pytest.helpers.temp_file(
        "fail_render.sls", state_file, base_env_state_tree_root_dir
    )
    with top_tempfile, state_tempfile:
        yield


@pytest.fixture(scope="class")
def state_tree_req_fail(base_env_state_tree_root_dir):
    top_file = """
    base:
      'localhost':
        - fail_req
      '127.0.0.1':
        - fail_req
    """
    state_file = """
    This has an invalid requisite:
      test.nop:
        - name: foo
        - require_in:
          - file.managed: invalid_requisite
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_state_tree_root_dir
    )
    state_tempfile = pytest.helpers.temp_file(
        "fail_req.sls", state_file, base_env_state_tree_root_dir
    )
    with top_tempfile, state_tempfile:
        yield


@pytest.fixture(scope="class")
def state_tree_structure_fail(base_env_state_tree_root_dir):
    top_file = """
    base:
      'localhost':
        - fail_structure
      '127.0.0.1':
        - fail_structure
    """
    state_file = """
    extend:
      Some file state:
        file:
            - name: /tmp/bar
            - contents: bar
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_state_tree_root_dir
    )
    state_tempfile = pytest.helpers.temp_file(
        "fail_structure.sls", state_file, base_env_state_tree_root_dir
    )
    with top_tempfile, state_tempfile:
        yield


@pytest.fixture(scope="class")
def state_tree_run_fail(base_env_state_tree_root_dir):
    top_file = """
    base:
      'localhost':
        - fail_run
      '127.0.0.1':
        - fail_run
    """
    state_file = """
    This file state fails:
      file.managed:
        - name: /tmp/non/ex/is/tent
        - makedirs: false
        - contents: foo
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_state_tree_root_dir
    )
    state_tempfile = pytest.helpers.temp_file(
        "fail_run.sls", state_file, base_env_state_tree_root_dir
    )
    with top_tempfile, state_tempfile:
        yield


@pytest.fixture(scope="class")
def pillar_tree_render_fail(base_env_pillar_tree_root_dir):
    top_file = """
    base:
      'localhost':
        - fail_render
      '127.0.0.1':
        - fail_render
    """
    pillar_file = r"""
    not_defined: {{ abc }}
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_pillar_tree_root_dir
    )
    pillar_tempfile = pytest.helpers.temp_file(
        "fail_render.sls", pillar_file, base_env_pillar_tree_root_dir
    )
    with top_tempfile, pillar_tempfile:
        yield


@pytest.mark.slow_test
def test_state_with_import(salt_ssh_cli, state_tree):
    """
    verify salt-ssh can use imported map files in states
    """
    ret = salt_ssh_cli.run("state.sls", "test")
    assert ret.returncode == 0
    assert ret.data


@pytest.mark.parametrize(
    "ssh_cmd",
    [
        "state.sls",
        "state.highstate",
        "state.apply",
        "state.show_top",
        "state.show_highstate",
        "state.show_low_sls",
        "state.show_lowstate",
        "state.sls_id",
        "state.show_sls",
        "state.top",
    ],
)
@pytest.mark.slow_test
def test_state_with_import_dir(salt_ssh_cli, state_tree_dir, ssh_cmd):
    """
    verify salt-ssh can use imported map files in states
    when the map files are in another directory outside of
    sls files importing them.
    """
    if ssh_cmd in ("state.sls", "state.show_low_sls", "state.show_sls"):
        ret = salt_ssh_cli.run("-w", "-t", ssh_cmd, "test")
    elif ssh_cmd == "state.top":
        ret = salt_ssh_cli.run("-w", "-t", ssh_cmd, "top.sls")
    elif ssh_cmd == "state.sls_id":
        ret = salt_ssh_cli.run("-w", "-t", ssh_cmd, "Ok with def", "test")
    else:
        ret = salt_ssh_cli.run("-w", "-t", ssh_cmd)
    assert ret.returncode == 0
    if ssh_cmd == "state.show_top":
        assert ret.data == {"base": ["test", "master_tops_test"]} or {"base": ["test"]}
    elif ssh_cmd in ("state.show_highstate", "state.show_sls"):
        assert ret.data == {
            "Ok with def": {
                "__sls__": "test",
                "__env__": "base",
                "test": ["succeed_without_changes", {"order": 10000}],
            }
        }
    elif ssh_cmd in ("state.show_low_sls", "state.show_lowstate", "state.show_sls"):
        assert ret.data == [
            {
                "state": "test",
                "name": "Ok with def",
                "__sls__": "test",
                "__env__": "base",
                "__id__": "Ok with def",
                "order": 10000,
                "fun": "succeed_without_changes",
            }
        ]
    else:
        assert ret.data["test_|-Ok with def_|-Ok with def_|-succeed_without_changes"][
            "result"
        ]
    assert ret.data


@pytest.fixture
def nested_state_tree(base_env_state_tree_root_dir, tmp_path):
    top_file = """
    base:
      'localhost':
        - basic
      '127.0.0.1':
        - basic
    """
    state_file = """
    /{}/file.txt:
      file.managed:
        - source: salt://foo/file.jinja
        - template: jinja
    """.format(
        tmp_path
    )
    file_jinja = """
    {% from 'foo/map.jinja' import comment %}{{ comment }}
    """
    map_file = """
    {% set comment = "blah blah" %}
    """
    statedir = base_env_state_tree_root_dir / "foo"
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_state_tree_root_dir
    )
    map_tempfile = pytest.helpers.temp_file("map.jinja", map_file, statedir)
    file_tempfile = pytest.helpers.temp_file("file.jinja", file_jinja, statedir)
    state_tempfile = pytest.helpers.temp_file("init.sls", state_file, statedir)

    with top_tempfile, map_tempfile, state_tempfile, file_tempfile:
        yield


@pytest.mark.slow_test
def test_state_with_import_from_dir(salt_ssh_cli, nested_state_tree):
    """
    verify salt-ssh can use imported map files in states
    """
    ret = salt_ssh_cli.run(
        "--extra-filerefs=salt://foo/map.jinja", "state.apply", "foo"
    )
    assert ret.returncode == 0
    assert ret.data


@pytest.mark.slow_test
def test_state_low(salt_ssh_cli):
    """
    test state.low with salt-ssh
    """
    ret = salt_ssh_cli.run(
        "state.low", '{"state": "cmd", "fun": "run", "name": "echo blah"}'
    )
    assert (
        json.loads(ret.stdout)["localhost"]["cmd_|-echo blah_|-echo blah_|-run"][
            "changes"
        ]["stdout"]
        == "blah"
    )


@pytest.mark.slow_test
def test_state_high(salt_ssh_cli):
    """
    test state.high with salt-ssh
    """
    ret = salt_ssh_cli.run("state.high", '{"echo blah": {"cmd": ["run"]}}')
    assert (
        json.loads(ret.stdout)["localhost"]["cmd_|-echo blah_|-echo blah_|-run"][
            "changes"
        ]["stdout"]
        == "blah"
    )


@pytest.mark.slow_test
@pytest.mark.usefixtures("state_tree_render_fail")
class TestRenderExceptionRetcode:
    """
    Verify salt-ssh fails with a retcode > 0 when a state rendering fails.
    """

    def test_retcode_state_sls_render_exception(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.sls", "fail_render")
        self._assert_ret(ret, EX_AGGREGATE)

    def test_retcode_state_highstate_render_exception(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.highstate")
        self._assert_ret(ret, EX_AGGREGATE)

    def test_retcode_state_sls_id_render_exception(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.sls_id", "foo", "fail_render")
        self._assert_ret(ret, EX_AGGREGATE)

    def test_retcode_state_show_sls_render_exception(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.show_sls", "fail_render")
        self._assert_ret(ret, EX_AGGREGATE)

    def test_retcode_state_show_low_sls_render_exception(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.show_low_sls", "fail_render")
        self._assert_ret(ret, EX_AGGREGATE)

    def test_retcode_state_show_highstate_render_exception(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.show_highstate")
        self._assert_ret(ret, EX_AGGREGATE)

    def test_retcode_state_show_lowstate_render_exception(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.show_lowstate")
        # state.show_lowstate exits with 0 for non-ssh as well
        self._assert_ret(ret, 0)

    def test_retcode_state_top_render_exception(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.top", "top.sls")
        self._assert_ret(ret, EX_AGGREGATE)

    def test_retcode_state_single_render_exception(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.single", "file")
        assert ret.returncode == EX_AGGREGATE
        assert isinstance(ret.data, str)
        assert "single() missing 1 required positional argument" in ret.data

    def _assert_ret(self, ret, retcode):
        assert ret.returncode == retcode
        assert isinstance(ret.data, list)
        assert ret.data
        assert isinstance(ret.data[0], str)
        assert ret.data[0].startswith(
            "Rendering SLS 'base:fail_render' failed: Jinja variable 'abc' is undefined;"
        )


@pytest.mark.slow_test
@pytest.mark.usefixtures("pillar_tree_render_fail")
class TestPillarRenderExceptionRetcode:
    """
    Verify salt-ssh fails with a retcode > 0 when a pillar rendering fails.
    """

    def test_retcode_state_sls_pillar_render_exception(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.sls", "basic")
        self._assert_ret(ret)

    def test_retcode_state_highstate_pillar_render_exception(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.highstate")
        self._assert_ret(ret)

    def test_retcode_state_sls_id_pillar_render_exception(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.sls_id", "foo", "basic")
        self._assert_ret(ret)

    def test_retcode_state_show_sls_pillar_render_exception(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.show_sls", "basic")
        self._assert_ret(ret)

    def test_retcode_state_show_low_sls_pillar_render_exception(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.show_low_sls", "basic")
        self._assert_ret(ret)

    def test_retcode_state_show_highstate_pillar_render_exception(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.show_highstate")
        self._assert_ret(ret)

    def test_retcode_state_show_lowstate_pillar_render_exception(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.show_lowstate")
        self._assert_ret(ret)

    def test_retcode_state_top_pillar_render_exception(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.top", "top.sls")
        self._assert_ret(ret)

    def _assert_ret(self, ret):
        assert ret.returncode == EX_AGGREGATE
        assert isinstance(ret.data, list)
        assert ret.data
        assert isinstance(ret.data[0], str)
        assert ret.data[0] == "Pillar failed to render with the following messages:"
        assert ret.data[1].startswith("Rendering SLS 'fail_render' failed.")


@pytest.mark.slow_test
@pytest.mark.usefixtures("state_tree_req_fail")
class TestStateReqFailRetcode:
    """
    Verify salt-ssh fails with a retcode > 0 when a highstate verification fails.
    ``state.show_highstate`` does not validate this.
    """

    def test_retcode_state_sls_invalid_requisite(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.sls", "fail_req")
        self._assert_ret(ret, EX_AGGREGATE)

    def test_retcode_state_highstate_invalid_requisite(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.highstate")
        self._assert_ret(ret, EX_AGGREGATE)

    def test_retcode_state_show_sls_invalid_requisite(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.show_sls", "fail_req")
        self._assert_ret(ret, EX_AGGREGATE)

    def test_retcode_state_show_low_sls_invalid_requisite(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.show_low_sls", "fail_req")
        self._assert_ret(ret, EX_AGGREGATE)

    def test_retcode_state_show_lowstate_invalid_requisite(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.show_lowstate")
        # state.show_lowstate exits with 0 for non-ssh as well
        self._assert_ret(ret, 0)

    def test_retcode_state_top_invalid_requisite(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.top", "top.sls")
        self._assert_ret(ret, EX_AGGREGATE)

    def _assert_ret(self, ret, retcode):
        assert ret.returncode == retcode
        assert isinstance(ret.data, list)
        assert ret.data
        assert isinstance(ret.data[0], str)
        assert ret.data[0].startswith(
            "Invalid requisite in require: file.managed for invalid_requisite"
        )


@pytest.mark.slow_test
@pytest.mark.usefixtures("state_tree_structure_fail")
class TestStateStructureFailRetcode:
    """
    Verify salt-ssh fails with a retcode > 0 when a highstate verification fails.
    This targets another step of the verification.
    ``state.sls_id`` does not seem to support extends.
    ``state.show_highstate`` does not validate this.
    """

    def test_retcode_state_sls_invalid_structure(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.sls", "fail_structure")
        self._assert_ret(ret, EX_AGGREGATE)

    def test_retcode_state_highstate_invalid_structure(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.highstate")
        self._assert_ret(ret, EX_AGGREGATE)

    def test_retcode_state_show_sls_invalid_structure(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.show_sls", "fail_structure")
        self._assert_ret(ret, EX_AGGREGATE)

    def test_retcode_state_show_low_sls_invalid_structure(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.show_low_sls", "fail_structure")
        self._assert_ret(ret, EX_AGGREGATE)

    def test_retcode_state_show_lowstate_invalid_structure(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.show_lowstate")
        # state.show_lowstate exits with 0 for non-ssh as well
        self._assert_ret(ret, 0)

    def test_retcode_state_top_invalid_structure(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.top", "top.sls")
        self._assert_ret(ret, EX_AGGREGATE)

    def _assert_ret(self, ret, retcode):
        assert ret.returncode == retcode
        assert isinstance(ret.data, list)
        assert ret.data
        assert isinstance(ret.data[0], str)
        assert ret.data[0].startswith(
            "Cannot extend ID 'Some file state' in 'base:fail_structure"
        )


@pytest.mark.slow_test
@pytest.mark.usefixtures("state_tree_run_fail")
class TestStateRunFailRetcode:
    """
    Verify salt-ssh passes on a failing retcode from state execution.
    """

    def test_retcode_state_sls_run_fail(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.sls", "fail_run")
        assert ret.returncode == EX_AGGREGATE

    def test_retcode_state_highstate_run_fail(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.highstate")
        assert ret.returncode == EX_AGGREGATE

    def test_retcode_state_sls_id_render_exception(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.sls_id", "This file state fails", "fail_run")
        assert ret.returncode == EX_AGGREGATE

    def test_retcode_state_top_run_fail(self, salt_ssh_cli):
        ret = salt_ssh_cli.run("state.top", "top.sls")
        assert ret.returncode == EX_AGGREGATE



"""
This tests policies that are defined in the giant dictionary in the LGPO module
"""
import pytest

import salt.modules.win_file as win_file
import salt.modules.win_lgpo as win_lgpo
import salt.utils.win_reg as win_reg

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.slow_test,
]


@pytest.fixture
def configure_loader_modules(tmp_path):
    cache_dir = tmp_path / "__test_admx_policy_cache_dir"
    cache_dir.mkdir(parents=True, exist_ok=True)
    return {
        win_lgpo: {
            "__salt__": {
                "file.file_exists": win_file.file_exists,
                "file.makedirs": win_file.makedirs_,
            },
            "__opts__": {
                "cachedir": str(cache_dir),
            },
            "__utils__": {
                "reg.set_value": win_reg.set_value,
                "reg.read_value": win_reg.read_value,
                "reg.delete_value": win_reg.delete_value,
            },
            "__context__": {},
        },
    }


@pytest.mark.destructive_test
def test_vuln_channel_allow():
    key = "SYSTEM\\CurrentControlSet\\Services\\Netlogon\\Parameters"
    vname = "VulnerableChannelAllowList"
    setting = "O:BAG:BAD:(A;;RC;;;BA)"
    win_reg.delete_value(hive="HKLM", key=key, vname=vname)
    try:
        win_lgpo.set_computer_policy(name=vname, setting=setting)
        result = win_reg.read_value(hive="HKLM", key=key, vname=vname)
        assert result["vdata"] == setting
    finally:
        win_reg.delete_value(hive="HKLM", key=key, vname=vname)


@pytest.mark.destructive_test
def test_vuln_channel_allow_not_defined():
    key = "SYSTEM\\CurrentControlSet\\Services\\Netlogon\\Parameters"
    vname = "VulnerableChannelAllowList"
    win_reg.set_value(hive="HKLM", key=key, vname=vname, vdata="junk", vtype="REG_SZ")
    try:
        win_lgpo.set_computer_policy(name=vname, setting="Not Defined")
        assert not win_reg.value_exists(hive="HKLM", key=key, vname=vname)
    finally:
        win_reg.delete_value(hive="HKLM", key=key, vname=vname)


@pytest.mark.destructive_test
def test_ldap_channel_binding_never():
    key = "SYSTEM\\CurrentControlSet\\Services\\NTDS\\Parameters"
    vname = "LdapEnforceChannelBinding"
    setting = "Never"
    win_reg.delete_value(hive="HKLM", key=key, vname=vname)
    try:
        win_lgpo.set_computer_policy(name=vname, setting=setting)
        result = win_reg.read_value(hive="HKLM", key=key, vname=vname)
        assert result["vdata"] == 0
    finally:

        win_reg.delete_value(hive="HKLM", key=key, vname=vname)


@pytest.mark.destructive_test
def test_ldap_channel_binding_when_supported():
    key = "SYSTEM\\CurrentControlSet\\Services\\NTDS\\Parameters"
    vname = "LdapEnforceChannelBinding"
    setting = "When supported"
    win_reg.delete_value(hive="HKLM", key=key, vname=vname)
    try:
        win_lgpo.set_computer_policy(name=vname, setting=setting)
        result = win_reg.read_value(hive="HKLM", key=key, vname=vname)
        assert result["vdata"] == 1
    finally:

        win_reg.delete_value(hive="HKLM", key=key, vname=vname)


@pytest.mark.destructive_test
def test_ldap_channel_binding_always():
    key = "SYSTEM\\CurrentControlSet\\Services\\NTDS\\Parameters"
    vname = "LdapEnforceChannelBinding"
    setting = "Always"
    win_reg.delete_value(hive="HKLM", key=key, vname=vname)
    try:
        win_lgpo.set_computer_policy(name=vname, setting=setting)
        result = win_reg.read_value(hive="HKLM", key=key, vname=vname)
        assert result["vdata"] == 2
    finally:

        win_reg.delete_value(hive="HKLM", key=key, vname=vname)


@pytest.mark.destructive_test
def test_ldap_channel_binding_not_defined():
    key = "SYSTEM\\CurrentControlSet\\Services\\NTDS\\Parameters"
    vname = "LdapEnforceChannelBinding"
    win_reg.set_value(hive="HKLM", key=key, vname=vname, vdata=1, vtype="REG_DWORD")
    try:
        win_lgpo.set_computer_policy(name=vname, setting="Not Defined")
        assert not win_reg.value_exists(hive="HKLM", key=key, vname=vname)
    finally:
        win_reg.delete_value(hive="HKLM", key=key, vname=vname)



"""
Manage etcd Keys
================

.. versionadded:: 2015.8.0

:depends:  - python-etcd or etcd3-py

This state module supports setting and removing keys from etcd.

Configuration
-------------

To work with an etcd server you must configure an etcd profile. The etcd config
can be set in either the Salt Minion configuration file or in pillar:

.. code-block:: yaml

    my_etd_config:
      etcd.host: 127.0.0.1
      etcd.port: 4001

It is technically possible to configure etcd without using a profile, but this
is not considered to be a best practice, especially when multiple etcd servers
or clusters are available.

.. code-block:: yaml

    etcd.host: 127.0.0.1
    etcd.port: 4001

In order to choose whether to use etcd API v2 or v3, you can put the following
configuration option in the same place as your etcd configuration.  This option
defaults to true, meaning you will use v2 unless you specify otherwise.

.. code-block:: yaml

    etcd.require_v2: True

When using API v3, there are some specific options available to be configured
within your etcd profile.  They are defaulted to the following...

.. code-block:: yaml

    etcd.encode_keys: False
    etcd.encode_values: True
    etcd.raw_keys: False
    etcd.raw_values: False
    etcd.unicode_errors: "surrogateescape"

``etcd.encode_keys`` indicates whether you want to pre-encode keys using msgpack before
adding them to etcd.

.. note::

    If you set ``etcd.encode_keys`` to ``True``, all recursive functionality will no longer work.
    This includes ``tree`` and ``ls`` and all other methods if you set ``recurse``/``recursive`` to ``True``.
    This is due to the fact that when encoding with msgpack, keys like ``/salt`` and ``/salt/stack`` will have
    differing byte prefixes, and etcd v3 searches recursively using prefixes.

``etcd.encode_values`` indicates whether you want to pre-encode values using msgpack before
adding them to etcd.  This defaults to ``True`` to avoid data loss on non-string values wherever possible.

``etcd.raw_keys`` determines whether you want the raw key or a string returned.

``etcd.raw_values`` determines whether you want the raw value or a string returned.

``etcd.unicode_errors`` determines what you policy to follow when there are encoding/decoding errors.

.. note::

    The etcd configuration can also be set in the Salt Master config file,
    but in order to use any etcd configurations defined in the Salt Master
    config, the :conf_master:`pillar_opts` must be set to ``True``.

    Be aware that setting ``pillar_opts`` to ``True`` has security implications
    as this makes all master configuration settings available in all minion's
    pillars.

Etcd profile configuration can be overridden using following arguments: ``host``,
``port``, ``username``, ``password``, ``ca``, ``client_key`` and ``client_cert``.
The v3 specific arguments can also be used for overriding if you are using v3.

.. code-block:: yaml

    my-value:
      etcd.set:
        - name: /path/to/key
        - value: value
        - host: 127.0.0.1
        - port: 2379
        - username: user
        - password: pass

Available Functions
-------------------

- ``set``

  This will set a value to a key in etcd. Changes will be returned if the key
  has been created or the value of the key has been updated. This
  means you can watch these states for changes.

  .. code-block:: yaml

      /foo/bar/baz:
        etcd.set:
          - value: foo
          - profile: my_etcd_config

- ``wait_set``

  Performs the same functionality as ``set`` but only if a watch requisite is ``True``.

  .. code-block:: yaml

      /some/file.txt:
        file.managed:
          - source: salt://file.txt

      /foo/bar/baz:
        etcd.wait_set:
          - value: foo
          - profile: my_etcd_config
          - watch:
            - file: /some/file.txt

- ``rm``

  This will delete a key from etcd. If the key exists then changes will be
  returned and thus you can watch for changes on the state, if the key does
  not exist then no changes will occur.

  .. code-block:: yaml

      /foo/bar/baz:
        etcd.rm:
          - profile: my_etcd_config

- ``wait_rm``

  Performs the same functionality as ``rm`` but only if a watch requisite is ``True``.

  .. code-block:: yaml

      /some/file.txt:
        file.managed:
          - source: salt://file.txt

      /foo/bar/baz:
        etcd.wait_rm:
          - profile: my_etcd_config
          - watch:
            - file: /some/file.txt
"""


# Define the module's virtual name
__virtualname__ = "etcd"

# Function aliases
__func_alias__ = {
    "set_": "set",
}

try:
    import salt.utils.etcd_util  # pylint: disable=W0611

    if salt.utils.etcd_util.HAS_ETCD_V2 or salt.utils.etcd_util.HAS_ETCD_V3:
        HAS_LIBS = True
    else:
        HAS_LIBS = False
except ImportError:
    HAS_LIBS = False

NO_PROFILE_MSG = "No profile found, using a profile is always recommended"


def __virtual__():
    """
    Only return if python-etcd is installed
    """
    if HAS_LIBS:
        return __virtualname__
    return (False, "Unable to import etcd_util")


def _etcd_action(*, action, key, profile, value=None, **kwargs):
    try:
        ret = __salt__["etcd.{}".format(action)](
            key=key, profile=profile, value=value, **kwargs
        )
    except Exception:  # pylint: disable=broad-except
        ret = None
    return ret


def set_(name, value, profile=None, **kwargs):
    """
    Set a key in etcd

    name
        The etcd key name, for example: ``/foo/bar/baz``.
    value
        The value the key should contain.

    profile
        Optional, defaults to ``None``. Sets the etcd profile to use which has
        been defined in the Salt Master config.

        .. code-block:: yaml

            my_etd_config:
              etcd.host: 127.0.0.1
              etcd.port: 4001

    """

    created = False

    rtn = {
        "name": name,
        "comment": "Key contains correct value",
        "result": True,
        "changes": {},
    }

    current = _etcd_action(action="get", key=name, profile=profile, **kwargs)

    if current is None and profile is None:
        rtn["comment"] = NO_PROFILE_MSG
        rtn["result"] = False
        return rtn

    if not current:
        created = True

    result = _etcd_action(
        action="set", key=name, value=value, profile=profile, **kwargs
    )

    if result and result != current:
        if created:
            rtn["comment"] = "New key created"
        else:
            rtn["comment"] = "Key value updated"
        rtn["changes"] = {name: value}

    return rtn


def wait_set(name, value, profile=None, **kwargs):
    """
    Set a key in etcd only if the watch statement calls it. This function is
    also aliased as ``wait_set``.

    name
        The etcd key name, for example: ``/foo/bar/baz``.
    value
        The value the key should contain.
    profile
        The etcd profile to use that has been configured on the Salt Master,
        this is optional and defaults to ``None``.

        .. code-block:: yaml

            my_etd_config:
              etcd.host: 127.0.0.1
              etcd.port: 4001

    """

    return {"name": name, "changes": {}, "result": True, "comment": ""}


def directory(name, profile=None, **kwargs):
    """
    Create a directory in etcd.

    name
        The etcd directory name, for example: ``/foo/bar/baz``.
    profile
        Optional, defaults to ``None``. Sets the etcd profile to use which has
        been defined in the Salt Master config.

        .. code-block:: yaml

            my_etd_config:
              etcd.host: 127.0.0.1
              etcd.port: 4001
    """

    created = False

    rtn = {"name": name, "comment": "Directory exists", "result": True, "changes": {}}

    current = _etcd_action(
        action="get", key=name, profile=profile, recurse=True, **kwargs
    )

    if current is None and profile is None:
        rtn["comment"] = NO_PROFILE_MSG
        rtn["result"] = False
        return rtn

    if not current:
        created = True

    result = __salt__["etcd.set"](name, None, directory=True, profile=profile, **kwargs)

    if result and result != current:
        if created:
            rtn["comment"] = "New directory created"
            rtn["changes"] = {name: "Created"}

    return rtn


def rm(name, recurse=False, profile=None, **kwargs):
    """
    Deletes a key from etcd

    name
        The etcd key name to remove, for example ``/foo/bar/baz``.

    recurse
        Optional, defaults to ``False``. If ``True`` performs a recursive delete.

    profile
        Optional, defaults to ``None``. Sets the etcd profile to use which has
        been defined in the Salt Master config.

        .. code-block:: yaml

            my_etd_config:
              etcd.host: 127.0.0.1
              etcd.port: 4001
    """

    rtn = {"name": name, "result": True, "changes": {}}

    current = _etcd_action(
        action="get", key=name, profile=profile, recurse=True, **kwargs
    )

    if current is None and profile is None:
        rtn["comment"] = NO_PROFILE_MSG
        rtn["result"] = False
        return rtn

    if not current:
        rtn["comment"] = "Key does not exist"
        return rtn

    if __salt__["etcd.rm"](name, recurse=recurse, profile=profile, **kwargs):
        rtn["comment"] = "Key removed"
        rtn["changes"] = {name: "Deleted"}
    else:
        rtn["comment"] = "Unable to remove key"

    return rtn


def wait_rm(name, recurse=False, profile=None, **kwargs):
    """
    Deletes a key from etcd only if the watch statement calls it.
    This function is also aliased as ``wait_rm``.

    name
        The etcd key name to remove, for example ``/foo/bar/baz``.
    recurse
        Optional, defaults to ``False``. If ``True`` performs a recursive
        delete, see: https://python-etcd.readthedocs.io/en/latest/#delete-a-key.
    profile
        Optional, defaults to ``None``. Sets the etcd profile to use which has
        been defined in the Salt Master config.

        .. code-block:: yaml

            my_etd_config:
              etcd.host: 127.0.0.1
              etcd.port: 4001
    """

    return {"name": name, "changes": {}, "result": True, "comment": ""}


def mod_watch(name, **kwargs):
    """
    The etcd watcher, called to invoke the watch command.
    When called, execute a etcd function based on a watch call requisite.

    .. note::
        This state exists to support special handling of the ``watch``
        :ref:`requisite <requisites>`. It should not be called directly.

        Parameters for this function should be set by the state being triggered.
    """

    # Watch to set etcd key
    if kwargs.get("sfun") in ["wait_set_key", "wait_set"]:
        return set_(name, kwargs.get("value"), profile=kwargs.get("profile"))

    # Watch to rm etcd key
    if kwargs.get("sfun") in ["wait_rm_key", "wait_rm"]:
        return rm(name, profile=kwargs.get("profile"))

    return {
        "name": name,
        "changes": {},
        "comment": (
            "etcd.{0[sfun]} does not work with the watch requisite, "
            "please use etcd.wait_set or etcd.wait_rm".format(kwargs)
        ),
        "result": False,
    }



"""
Tests for salt.utils.yamlencoding
"""


import salt.utils.yaml
import salt.utils.yamlencoding
from tests.support.unit import TestCase


class YamlEncodingTestCase(TestCase):
    def test_yaml_dquote(self):
        for teststr in (r'"\ []{}"',):
            self.assertEqual(
                teststr,
                salt.utils.yaml.safe_load(salt.utils.yamlencoding.yaml_dquote(teststr)),
            )

    def test_yaml_dquote_doesNotAddNewLines(self):
        teststr = '"' * 100
        self.assertNotIn("\n", salt.utils.yamlencoding.yaml_dquote(teststr))

    def test_yaml_squote(self):
        ret = salt.utils.yamlencoding.yaml_squote(r'"')
        self.assertEqual(ret, r"""'"'""")

    def test_yaml_squote_doesNotAddNewLines(self):
        teststr = "'" * 100
        self.assertNotIn("\n", salt.utils.yamlencoding.yaml_squote(teststr))

    def test_yaml_encode(self):
        for testobj in (
            None,
            True,
            False,
            "[7, 5]",
            '"monkey"',
            5,
            7.5,
            "2014-06-02 15:30:29.7",
        ):
            self.assertEqual(
                testobj,
                salt.utils.yaml.safe_load(salt.utils.yamlencoding.yaml_encode(testobj)),
            )

        for testobj in ({}, [], set()):
            self.assertRaises(TypeError, salt.utils.yamlencoding.yaml_encode, testobj)



import sys

import pytest

from tests.support.helpers import VirtualEnv


@pytest.mark.parametrize(
    "pip_version",
    (
        pytest.param(
            "pip==9.0.3",
            marks=pytest.mark.skipif(
                sys.version_info >= (3, 10),
                reason="'pip==9.0.3' is not available on Py >= 3.10",
            ),
        ),
        "pip<20.0",
        "pip<21.0",
        "pip>=21.0",
    ),
)
@pytest.mark.requires_network
@pytest.mark.slow_test
def test_list_available_packages(modules, pip_version, tmp_path):
    with VirtualEnv(venv_dir=tmp_path, pip_requirement=pip_version) as virtualenv:
        virtualenv.install("-U", pip_version)
        package_name = "pep8"
        available_versions = modules.pip.list_all_versions(
            package_name, bin_env=str(virtualenv.venv_bin_dir)
        )
        assert available_versions


@pytest.mark.parametrize(
    "pip_version",
    (
        "pip==9.0.3",
        "pip<20.0",
        "pip<21.0",
        "pip>=21.0",
    ),
)
def test_list_available_packages_with_index_url(modules, pip_version, tmp_path):
    if sys.version_info < (3, 6) and pip_version == "pip>=21.0":
        pytest.skip("{} is not available on Py3.5".format(pip_version))
    if sys.version_info >= (3, 10) and pip_version == "pip==9.0.3":
        pytest.skip("{} is not available on Py3.10".format(pip_version))
    with VirtualEnv(venv_dir=tmp_path, pip_requirement=pip_version) as virtualenv:
        virtualenv.install("-U", pip_version)
        package_name = "pep8"
        available_versions = modules.pip.list_all_versions(
            package_name,
            bin_env=str(virtualenv.venv_bin_dir),
            index_url="https://pypi.python.org/simple",
        )
        assert available_versions



from sys import platform

import pytest

pytestmark = [
    pytest.mark.skip_on_windows,
]


def test_salt_cmd_run(salt_cli, salt_minion):
    """
    Test salt cmd.run 'ipconfig' or 'ls -lah /'
    """
    ret = None
    if platform.startswith("win"):
        ret = salt_cli.run("cmd.run", "ipconfig", minion_tgt=salt_minion.id)
    else:
        ret = salt_cli.run("cmd.run", "ls -lah /", minion_tgt=salt_minion.id)
    assert ret
    assert ret.stdout


def test_salt_list_users(salt_cli, salt_minion):
    """
    Test salt user.list_users
    """
    ret = salt_cli.run("user.list_users", minion_tgt=salt_minion.id)
    if platform.startswith("win"):
        assert "Administrator" in ret.stdout
    else:
        assert "root" in ret.stdout



"""
unit tests for the git_pillar runner
"""


import logging

import pytest

import salt.runners.git_pillar as git_pillar
import salt.utils.files
import salt.utils.gitfs
from tests.support.gitfs import _OPTS
from tests.support.mock import patch

log = logging.getLogger(__name__)


@pytest.fixture
def cachedir(tmp_path):
    return tmp_path / "cache"


@pytest.fixture
def configure_loader_modules(cachedir):
    opts = _OPTS.copy()
    opts["cachedir"] = str(cachedir)
    opts["verified_git_pillar_provider"] = "gitfoo"
    opts["ext_pillar"] = [
        {
            "git": [
                "master https://someurl/some",
                {"dev https://otherurl/other": [{"name": "somename"}]},
            ]
        }
    ]
    return {git_pillar: {"__opts__": opts}}


def test_update():
    """
    test git_pillar.update
    """

    class MockGitProvider(
        salt.utils.gitfs.GitProvider
    ):  # pylint: disable=abstract-method
        def init_remote(self):
            new = False
            self.repo = True
            return new

        def fetch(self):
            return True

        def clear_lock(self, lock_type="update"):
            pass  # return success, failed

    git_providers = {"gitfoo": MockGitProvider}

    repo1 = {"master https://someurl/some": True}
    repo2 = {"dev https://otherurl/other": True}
    all_repos = {
        "master https://someurl/some": True,
        "dev https://otherurl/other": True,
    }
    with patch.object(salt.utils.gitfs, "GIT_PROVIDERS", git_providers):
        assert git_pillar.update() == all_repos
        assert git_pillar.update(branch="master") == repo1
        assert git_pillar.update(branch="dev") == repo2
        assert git_pillar.update(repo="somename") == repo2



"""
tests.unit.runners.test_asam
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Unit tests for the asam runner
"""

import logging

import pytest

import salt.runners.asam as asam
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    opts = {
        "asam": {
            "prov1.domain.com": {
                "username": "TheUsername",
                "password": "ThePassword",
            }
        }
    }
    return {asam: {"__opts__": opts}}


def test_add_platform():
    parse_html_content = MagicMock()
    get_platform_set_name = MagicMock(return_value="plat-foo")
    requests_mock = MagicMock()

    # remove_platform
    with patch("salt.runners.asam._parse_html_content", parse_html_content), patch(
        "salt.runners.asam._get_platformset_name", get_platform_set_name
    ), patch("salt.runners.asam.requests.post", requests_mock):
        asam.add_platform("plat-foo-2", "plat-foo", "prov1.domain.com")

    requests_mock.assert_called_with(
        "https://prov1.domain.com:3451/config/PlatformSetConfig.html",
        auth=("TheUsername", "ThePassword"),
        data={"manual": "false"},
        verify=True,
    )


def test_remove_platform():
    parse_html_content = MagicMock()
    get_platform_set_name = MagicMock(return_value="plat-foo")
    requests_mock = MagicMock()

    # remove_platform
    with patch("salt.runners.asam._parse_html_content", parse_html_content), patch(
        "salt.runners.asam._get_platformset_name", get_platform_set_name
    ), patch("salt.runners.asam.requests.post", requests_mock):
        asam.remove_platform("plat-foo", "prov1.domain.com")

    requests_mock.assert_called_with(
        "https://prov1.domain.com:3451/config/PlatformConfig.html",
        auth=("TheUsername", "ThePassword"),
        data={
            "manual": "false",
            "platformName": "plat-foo",
            "platformSetName": "plat-foo",
            "postType": "platformRemove",
            "Submit": "Yes",
        },
        verify=True,
    )


def test_list_platforms():
    parse_html_content = MagicMock()
    get_platforms = MagicMock(return_value=["plat-foo", "plat-bar"])
    requests_mock = MagicMock()

    # remove_platform
    with patch("salt.runners.asam._parse_html_content", parse_html_content), patch(
        "salt.runners.asam._get_platforms", get_platforms
    ), patch("salt.runners.asam.requests.post", requests_mock):
        asam.list_platforms("prov1.domain.com")

    requests_mock.assert_called_with(
        "https://prov1.domain.com:3451/config/PlatformConfig.html",
        auth=("TheUsername", "ThePassword"),
        data={"manual": "false"},
        verify=True,
    )


def test_list_platform_sets():
    parse_html_content = MagicMock()
    get_platform_sets = MagicMock(return_value=["plat-foo", "plat-bar"])
    requests_mock = MagicMock()

    # remove_platform
    with patch("salt.runners.asam._parse_html_content", parse_html_content), patch(
        "salt.runners.asam._get_platforms", get_platform_sets
    ), patch("salt.runners.asam.requests.post", requests_mock):
        asam.list_platform_sets("prov1.domain.com")

    requests_mock.assert_called_with(
        "https://prov1.domain.com:3451/config/PlatformSetConfig.html",
        auth=("TheUsername", "ThePassword"),
        data={"manual": "false"},
        verify=True,
    )



import os.path

import pytest

from salt.fileserver.gitfs import PER_REMOTE_ONLY, PER_REMOTE_OVERRIDES
from salt.utils.gitfs import GitFS, GitPython, Pygit2
from salt.utils.immutabletypes import ImmutableDict, ImmutableList

pytestmark = [
    pytest.mark.slow_test,
]


try:
    import git  # pylint: disable=unused-import

    HAS_GITPYTHON = True
except ImportError:
    HAS_GITPYTHON = False


try:
    import pygit2  # pylint: disable=unused-import

    HAS_PYGIT2 = True
except ImportError:
    HAS_PYGIT2 = False


skipif_no_gitpython = pytest.mark.skipif(not HAS_GITPYTHON, reason="Missing gitpython")
skipif_no_pygit2 = pytest.mark.skipif(not HAS_PYGIT2, reason="Missing pygit2")


@pytest.fixture
def gitfs_opts(salt_factories, tmp_path):
    config_defaults = {"cachedir": str(tmp_path)}
    factory = salt_factories.salt_master_daemon(
        "gitfs-functional-master", defaults=config_defaults
    )
    config_defaults = dict(factory.config)
    for key, item in config_defaults.items():
        if isinstance(item, ImmutableDict):
            config_defaults[key] = dict(item)
        elif isinstance(item, ImmutableList):
            config_defaults[key] = list(item)
    return config_defaults


@pytest.fixture
def gitpython_gitfs_opts(gitfs_opts):
    gitfs_opts["verified_gitfs_provider"] = "gitpython"
    GitFS.instance_map.clear()  # wipe instance_map object map for clean run
    return gitfs_opts


@pytest.fixture
def pygit2_gitfs_opts(gitfs_opts):
    gitfs_opts["verified_gitfs_provider"] = "pygit2"
    GitFS.instance_map.clear()  # wipe instance_map object map for clean run
    return gitfs_opts


def _get_gitfs(opts, *remotes):
    return GitFS(
        opts,
        remotes,
        per_remote_overrides=PER_REMOTE_OVERRIDES,
        per_remote_only=PER_REMOTE_ONLY,
    )


def _test_gitfs_simple(gitfs_opts):
    g = _get_gitfs(
        gitfs_opts,
        {"https://github.com/saltstack/salt-test-pillar-gitfs.git": [{"name": "bob"}]},
    )
    g.fetch_remotes()
    assert len(g.remotes) == 1
    assert set(g.file_list({"saltenv": "main"})) == {".gitignore", "README.md"}


@skipif_no_gitpython
def test_gitpython_gitfs_simple(gitpython_gitfs_opts):
    _test_gitfs_simple(gitpython_gitfs_opts)


@skipif_no_pygit2
def test_pygit2_gitfs_simple(pygit2_gitfs_opts):
    _test_gitfs_simple(pygit2_gitfs_opts)


def _test_gitfs_simple_base(gitfs_opts):
    g = _get_gitfs(
        gitfs_opts, "https://github.com/saltstack/salt-test-pillar-gitfs.git"
    )
    g.fetch_remotes()
    assert len(g.remotes) == 1
    assert set(g.file_list({"saltenv": "base"})) == {
        ".gitignore",
        "README.md",
        "file.sls",
        "top.sls",
    }


@skipif_no_gitpython
def test_gitpython_gitfs_simple_base(gitpython_gitfs_opts):
    _test_gitfs_simple_base(gitpython_gitfs_opts)


@skipif_no_pygit2
def test_pygit2_gitfs_simple_base(pygit2_gitfs_opts):
    _test_gitfs_simple_base(pygit2_gitfs_opts)


@skipif_no_gitpython
def test_gitpython_gitfs_provider(gitpython_gitfs_opts):
    g = _get_gitfs(
        gitpython_gitfs_opts, "https://github.com/saltstack/salt-test-pillar-gitfs.git"
    )
    assert len(g.remotes) == 1
    assert g.provider == "gitpython"
    assert isinstance(g.remotes[0], GitPython)


@skipif_no_pygit2
def test_pygit2_gitfs_provider(pygit2_gitfs_opts):
    g = _get_gitfs(
        pygit2_gitfs_opts, "https://github.com/saltstack/salt-test-pillar-gitfs.git"
    )
    assert len(g.remotes) == 1
    assert g.provider == "pygit2"
    assert isinstance(g.remotes[0], Pygit2)


def _test_gitfs_minion(gitfs_opts):
    gitfs_opts["__role"] = "minion"
    g = _get_gitfs(
        gitfs_opts, "https://github.com/saltstack/salt-test-pillar-gitfs.git"
    )
    g.fetch_remotes()
    assert len(g.remotes) == 1
    assert set(g.file_list({"saltenv": "base"})) == {
        ".gitignore",
        "README.md",
        "file.sls",
        "top.sls",
    }
    assert set(g.file_list({"saltenv": "main"})) == {".gitignore", "README.md"}


@skipif_no_gitpython
def test_gitpython_gitfs_minion(gitpython_gitfs_opts):
    _test_gitfs_minion(gitpython_gitfs_opts)


@skipif_no_pygit2
def test_pygit2_gitfs_minion(pygit2_gitfs_opts):
    _test_gitfs_minion(pygit2_gitfs_opts)


def _test_fetch_request_with_mountpoint(opts):
    mpoint = [{"mountpoint": "salt/m"}]
    p = _get_gitfs(
        opts,
        {"https://github.com/saltstack/salt-test-pillar-gitfs.git": mpoint},
    )
    p.fetch_remotes()
    assert len(p.remotes) == 1
    repo = p.remotes[0]
    assert repo.mountpoint("testmount") == "salt/m"
    assert set(p.file_list({"saltenv": "testmount"})) == {
        "salt/m/test_dir1/testfile3",
        "salt/m/test_dir1/test_dir2/testfile2",
        "salt/m/.gitignore",
        "salt/m/README.md",
        "salt/m/test_dir1/test_dir2/testfile1",
    }


@skipif_no_gitpython
def test_gitpython_fetch_request_with_mountpoint(gitpython_gitfs_opts):
    _test_fetch_request_with_mountpoint(gitpython_gitfs_opts)


@skipif_no_pygit2
def test_pygit2_fetch_request_with_mountpoint(pygit2_gitfs_opts):
    _test_fetch_request_with_mountpoint(pygit2_gitfs_opts)


def _test_name(opts):
    p = _get_gitfs(
        opts,
        {
            "https://github.com/saltstack/salt-test-pillar-gitfs.git": [
                {"name": "name1"}
            ]
        },
        {
            "https://github.com/saltstack/salt-test-pillar-gitfs.git": [
                {"name": "name2"}
            ]
        },
    )
    p.fetch_remotes()
    assert len(p.remotes) == 2
    repo = p.remotes[0]
    repo2 = p.remotes[1]
    assert repo.get_cache_basehash() == "name1"
    assert repo2.get_cache_basehash() == "name2"


@skipif_no_gitpython
def test_gitpython_name(gitpython_gitfs_opts):
    _test_name(gitpython_gitfs_opts)


@skipif_no_pygit2
def test_pygit2_name(pygit2_gitfs_opts):
    _test_name(pygit2_gitfs_opts)


def _test_remote_map(opts):
    p = _get_gitfs(
        opts,
        "https://github.com/saltstack/salt-test-pillar-gitfs.git",
    )
    p.fetch_remotes()
    assert len(p.remotes) == 1
    assert os.path.isfile(os.path.join(opts["cachedir"], "gitfs", "remote_map.txt"))


@skipif_no_gitpython
def test_gitpython_remote_map(gitpython_gitfs_opts):
    _test_remote_map(gitpython_gitfs_opts)


@skipif_no_pygit2
def test_pygit2_remote_map(pygit2_gitfs_opts):
    _test_remote_map(pygit2_gitfs_opts)



"""
Tests for salt.utils.jinja
"""

import os

import pytest

# dateutils is needed so that the strftime jinja filter is loaded
import salt.utils.dateutils  # pylint: disable=unused-import
import salt.utils.files  # pylint: disable=unused-import
import salt.utils.json  # pylint: disable=unused-import
import salt.utils.stringutils  # pylint: disable=unused-import
import salt.utils.yaml  # pylint: disable=unused-import
from salt.utils.templates import render_jinja_tmpl


@pytest.fixture
def minion_opts(tmp_path, minion_opts):
    minion_opts.update(
        {
            "cachedir": str(tmp_path / "jinja-template-cache"),
            "file_buffer_size": 1048576,
            "file_client": "local",
            "file_ignore_regex": None,
            "file_ignore_glob": None,
            "file_roots": {"test": [str(tmp_path / "templates")]},
            "pillar_roots": {"test": [str(tmp_path / "templates")]},
            "fileserver_backend": ["roots"],
            "hash_type": "md5",
            "extension_modules": os.path.join(
                os.path.dirname(os.path.abspath(__file__)), "extmods"
            ),
            "jinja_env": {"line_comment_prefix": "##", "line_statement_prefix": "%"},
        }
    )
    return minion_opts


@pytest.fixture
def local_salt():
    return {
        "myvar": "zero",
        "mylist": [0, 1, 2, 3],
    }


def test_comment_prefix(minion_opts, local_salt):

    template = """
        %- set myvar = 'one'
        ## ignored comment 1
        {{- myvar -}}
        {%- set myvar = 'two' %} ## ignored comment 2
        {{- myvar }} ## ignored comment 3
        %- if myvar == 'two':
        %- set myvar = 'three'
        %- endif
        {{- myvar -}}
        """
    rendered = render_jinja_tmpl(
        template, dict(opts=minion_opts, saltenv="test", salt=local_salt)
    )
    assert rendered == "onetwothree"


def test_statement_prefix(minion_opts, local_salt):

    template = """
        {%- set mylist = ['1', '2', '3'] %}
        %- set mylist = ['one', 'two', 'three']
        %- for item in mylist:
        {{- item }}
        %- endfor
        """
    rendered = render_jinja_tmpl(
        template, dict(opts=minion_opts, saltenv="test", salt=local_salt)
    )
    assert rendered == "onetwothree"
